import json
import os
import sys
from pathlib import Path

import numpy as np

import pygama.lgdo as lgdo
from pygama.dsp import build_processing_chain as bpc
from pygama.raw.buffer_processor.lh5_buffer_processor import lh5_buffer_processor
from pygama.raw.build_raw import build_raw
from pygama.raw.fc.fc_event_decoder import fc_decoded_values

# skip compression in build_raw
fc_decoded_values["waveform"].pop("compression", None)

config_dir = Path(__file__).parent / "test_buffer_processor_configs"


# check that packet indexes match in verification test
def test_lh5_buffer_processor_packet_ids(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca")
    raw_file = "/tmp/L200-comm-20220519-phy-geds.lh5"
    proc_out_spec = f"{config_dir}/lh5_buffer_processor_config.json"
    raw_out_spec = f"{config_dir}/raw_out_spec_no_proc.json"

    processed_file = raw_file.replace(
        "L200-comm-20220519-phy-geds.lh5", "L200-comm-20220519-phy-geds_proc.lh5"
    )
    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)

    lh5_buffer_processor(
        lh5_raw_file_in=raw_file, overwrite=True, out_spec=proc_out_spec
    )

    sto = lgdo.LH5Store()

    raw_group = "ORFlashCamADCWaveform"
    raw_packet_ids, _ = sto.read_object(str(raw_group) + "/packet_id", raw_file)
    processed_packet_ids, _ = sto.read_object(
        str(raw_group) + "/packet_id", processed_file
    )

    assert np.array_equal(raw_packet_ids.nda, processed_packet_ids.nda)

    processed_presummed_wfs, _ = sto.read_object(
        str(raw_group) + "/presummed_waveform/values", processed_file
    )
    raw_wfs, _ = sto.read_object(str(raw_group) + "/waveform/values", raw_file)
    assert processed_presummed_wfs.nda[0][0] == np.sum(raw_wfs.nda[0][:4])


# check that packet indexes match in verification test
def test_lh5_buffer_processor_waveform_lengths(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio")
    raw_file = daq_file.replace(
        "L200-comm-20211130-phy-spms.fcio", "L200-comm-20211130-phy-spms.lh5"
    )
    processed_file = raw_file.replace(
        "L200-comm-20211130-phy-spms.lh5", "L200-comm-20211130-phy-spms_proc.lh5"
    )

    out_spec = {
        "FCEventDecoder": {
            "ch{key}": {
                "key_list": [[0, 6]],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 1000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": ["presum_rate", "presummed_waveform"],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/16, period=waveform.period*16, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            }
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                    },
                },
            }
        }
    }

    copy_out_spec = {
        "FCEventDecoder": {
            "ch{key}": {
                "key_list": [[0, 6]],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            }
        }
    }

    build_raw(in_stream=daq_file, out_spec=copy_out_spec, overwrite=True)
    lh5_buffer_processor(
        lh5_raw_file_in=raw_file,
        overwrite=True,
        out_spec=out_spec,
        proc_file_name=processed_file,
    )

    lh5_tables = lgdo.ls(raw_file)
    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for i, tb in enumerate(lh5_tables):
        if "raw" not in tb and lgdo.ls(raw_file, f"{tb}/raw"):
            lh5_tables[i] = f"{tb}/raw"
        elif not lgdo.ls(raw_file, tb):
            del lh5_tables[i]

    # Grab the proc_spec after keylist expansion from build_raw
    proc_spec = {}
    for key in out_spec["FCEventDecoder"].keys():
        if "proc_spec" in out_spec["FCEventDecoder"][key].keys():
            proc_spec[key] = out_spec["FCEventDecoder"][key].pop("proc_spec")

    jsonfile = proc_spec

    # Read in the presummed rate from the config file to modify the clock rate later
    presum_rate_string = jsonfile["ch0"]["dsp_config"]["processors"][
        "presum_rate, presummed_waveform"
    ]["args"][3]
    presum_rate_start_idx = presum_rate_string.find("/") + 1
    presum_rate_end_idx = presum_rate_string.find(",")
    presum_rate = int(presum_rate_string[presum_rate_start_idx:presum_rate_end_idx])

    # This needs to be overwritten with the correct windowing values set in buffer_processor.py
    window_start_index = jsonfile["ch0"]["window"][1]
    window_end_index = jsonfile["ch0"]["window"][2]

    sto = lgdo.LH5Store()

    for raw_group in lh5_tables:
        raw_packet_waveform_values = sto.read_object(
            str(raw_group) + "/waveform/values", raw_file
        )
        presummed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/presummed_waveform/values", processed_file
        )
        windowed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/windowed_waveform/values", processed_file
        )

        # Check that the lengths of the waveforms match what we expect
        assert len(raw_packet_waveform_values[0].nda[0]) == presum_rate * len(
            presummed_packet_waveform_values[0].nda[0]
        )
        assert isinstance(presummed_packet_waveform_values[0].nda[0][0], np.uint32)
        assert len(raw_packet_waveform_values[0].nda[0]) == len(
            windowed_packet_waveform_values[0].nda[0]
        ) + np.abs(window_start_index) + np.abs(window_end_index)
        assert isinstance(windowed_packet_waveform_values[0].nda[0][0], np.uint16)

        raw_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/waveform/t0", raw_file
        )
        raw_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/waveform/dt", raw_file
        )

        windowed_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/windowed_waveform/t0", processed_file
        )
        presummed_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/presummed_waveform/t0", processed_file
        )

        # Check that the t0s match what we expect, with the correct units
        assert np.array_equal(
            raw_packet_waveform_t0s.nda,
            windowed_packet_waveform_t0s.nda
            - window_start_index * raw_packet_waveform_dts.nda,
        )
        assert (
            windowed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )
        assert np.array_equal(
            raw_packet_waveform_t0s.nda, presummed_packet_waveform_t0s.nda
        )
        assert (
            presummed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )

        presummed_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/presummed_waveform/dt", processed_file
        )

        # Check that the dts match what we expect, with the correct units
        assert np.array_equal(
            raw_packet_waveform_dts.nda, presummed_packet_waveform_dts.nda / presum_rate
        )


def test_lh5_buffer_processor_file_size_decrease(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("orca/sis3316/coherent-run1141-bkg.orca")
    raw_file = daq_file.replace("coherent-run1141-bkg.orca", "coherent-run1141-bkg.lh5")

    processed_file = raw_file.replace(
        "coherent-run1141-bkg.lh5", "coherent-run1141-bkg_proc.lh5"
    )

    raw_out_spec = {
        "ORSIS3316WaveformDecoder": {
            "Card1": {"key_list": [48], "out_stream": raw_file, "out_name": "raw"}
        }
    }

    proc_out_spec = {
        "ORSIS3316WaveformDecoder": {
            "Card1": {
                "key_list": [48],
                "out_stream": processed_file,
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 1000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": ["presum_rate", "presummed_waveform"],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/4, period=waveform.period*4, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            }
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                    },
                },
            }
        }
    }

    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)

    lh5_buffer_processor(
        lh5_raw_file_in=raw_file,
        overwrite=True,
        out_spec=proc_out_spec,
        proc_file_name=processed_file,
    )

    lh5_tables = lgdo.ls(raw_file)
    for i, tb in enumerate(lh5_tables):
        if "raw" not in tb and lgdo.ls(raw_file, f"{tb}/raw"):
            lh5_tables[i] = f"{tb}/raw"
        elif not lgdo.ls(raw_file, tb):
            del lh5_tables[i]
    sto = lgdo.LH5Store()

    wf_size = 0

    for raw_group in lh5_tables:
        wf_size += sys.getsizeof(
            sto.read_object(str(raw_group) + "/waveform/values", raw_file)[0].nda
        )

        # Make sure that we are actually processing the waveforms
        raw_packet_waveform_values = sto.read_object(
            str(raw_group) + "/waveform/values", raw_file
        )
        presummed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/presummed_waveform/values", processed_file
        )
        windowed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/windowed_waveform/values", processed_file
        )

        # Check that the lengths of the waveforms match what we expect
        assert len(raw_packet_waveform_values[0].nda[0]) == 4 * len(
            presummed_packet_waveform_values[0].nda[0]
        )
        assert isinstance(presummed_packet_waveform_values[0].nda[0][0], np.uint32)
        assert len(raw_packet_waveform_values[0].nda[0]) == len(
            windowed_packet_waveform_values[0].nda[0]
        ) + 1000 + np.abs(-1000)
        assert isinstance(windowed_packet_waveform_values[0].nda[0][0], np.uint16)

    # Make sure we are taking up less space than a file that has two copies of the waveform table in it
    assert os.path.getsize(processed_file) < os.path.getsize(raw_file) + wf_size


# check that packet indexes match in verification test on file that has both spms and geds
def test_lh5_buffer_processor_separate_name_tables(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio")
    raw_file = daq_file.replace(
        "L200-comm-20211130-phy-spms.fcio", " L200-comm-fake-geds-and-spms.lh5"
    )
    processed_file = raw_file.replace(
        "L200-comm-fake-geds-and-spms.lh5", "L200-comm-fake-geds-and-spms_proc.lh5"
    )

    raw_out_spec = {
        "FCEventDecoder": {
            "geds": {
                "key_list": [[0, 3]],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
            "spms": {
                "key_list": [[3, 6]],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
        }
    }

    proc_out_spec = {
        "FCEventDecoder": {
            "geds": {
                "key_list": [[0, 3]],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 2000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": ["presum_rate", "presummed_waveform"],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/8, period=waveform.period*8, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            }
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                    },
                },
            },
            "spms": {
                "key_list": [[3, 6]],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 1000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": ["presum_rate", "presummed_waveform"],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/4, period=waveform.period*4, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            }
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                    },
                },
            },
        }
    }

    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)
    lh5_buffer_processor(
        lh5_raw_file_in=raw_file,
        overwrite=True,
        out_spec=proc_out_spec,
        proc_file_name=processed_file,
    )

    lh5_tables = lgdo.ls(raw_file)
    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for i, tb in enumerate(lh5_tables):
        if "raw" not in tb and lgdo.ls(raw_file, f"{tb}/raw"):
            lh5_tables[i] = f"{tb}/raw"
        elif not lgdo.ls(raw_file, tb):
            del lh5_tables[i]

    # Grab the proc_spec after keylist expansion from build_raw
    proc_spec = {}
    for key in proc_out_spec["FCEventDecoder"].keys():
        if "proc_spec" in proc_out_spec["FCEventDecoder"][key].keys():
            proc_spec[key] = proc_out_spec["FCEventDecoder"][key].pop("proc_spec")

    jsonfile = proc_spec

    sto = lgdo.LH5Store()

    for raw_group in lh5_tables:
        # First, check the packet ids
        raw_packet_ids, _ = sto.read_object(str(raw_group) + "/packet_id", raw_file)
        processed_packet_ids, _ = sto.read_object(
            str(raw_group) + "/packet_id", processed_file
        )

        assert np.array_equal(raw_packet_ids.nda, processed_packet_ids.nda)

        # Read in the presummed rate from the config file to modify the clock rate later
        group_name = raw_group.split("/raw")[0]
        presum_rate_string = jsonfile[group_name]["dsp_config"]["processors"][
            "presum_rate, presummed_waveform"
        ]["args"][3]
        presum_rate_start_idx = presum_rate_string.find("/") + 1
        presum_rate_end_idx = presum_rate_string.find(",")
        presum_rate = int(presum_rate_string[presum_rate_start_idx:presum_rate_end_idx])

        # This needs to be overwritten with the correct windowing values set in buffer_processor.py
        window_start_index = int(jsonfile[group_name]["window"][1])
        window_end_index = int(jsonfile[group_name]["window"][2])

        raw_packet_waveform_values = sto.read_object(
            str(raw_group) + "/waveform/values", raw_file
        )
        presummed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/presummed_waveform/values", processed_file
        )
        windowed_packet_waveform_values = sto.read_object(
            str(raw_group) + "/windowed_waveform/values", processed_file
        )

        # Check that the lengths of the waveforms match what we expect
        assert len(raw_packet_waveform_values[0].nda[0]) == presum_rate * len(
            presummed_packet_waveform_values[0].nda[0]
        )
        assert isinstance(presummed_packet_waveform_values[0].nda[0][0], np.uint32)
        assert len(raw_packet_waveform_values[0].nda[0]) == len(
            windowed_packet_waveform_values[0].nda[0]
        ) + np.abs(window_start_index) + np.abs(window_end_index)
        assert isinstance(windowed_packet_waveform_values[0].nda[0][0], np.uint16)

        raw_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/waveform/t0", raw_file
        )
        raw_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/waveform/dt", raw_file
        )

        windowed_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/windowed_waveform/t0", processed_file
        )
        presummed_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/presummed_waveform/t0", processed_file
        )

        # Check that the t0s match what we expect, with the correct units
        assert (
            raw_packet_waveform_t0s.nda[0]
            == windowed_packet_waveform_t0s.nda[0]
            - raw_packet_waveform_dts.nda[0] * window_start_index
        )
        assert (
            windowed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )
        assert raw_packet_waveform_t0s.nda[0] == presummed_packet_waveform_t0s.nda[0]
        assert (
            presummed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )

        presummed_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/presummed_waveform/dt", processed_file
        )

        # Check that the dts match what we expect, with the correct units
        assert (
            raw_packet_waveform_dts.nda[0]
            == presummed_packet_waveform_dts.nda[0] / presum_rate
        )


def test_raw_geds_no_proc_spms(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio")
    raw_file = daq_file.replace(
        "L200-comm-20211130-phy-spms.fcio", "L200-comm-test-pass.lh5"
    )
    processed_file = raw_file.replace(
        "L200-comm-test-pass.lh5", "L200-comm-test-pass_proc.lh5"
    )

    raw_out_spec = {
        "FCEventDecoder": {
            "geds": {
                "key_list": [[0, 1]],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
            "spms": {
                "key_list": [[3, 4]],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
        }
    }

    proc_out_spec = {
        "FCEventDecoder": {
            "geds": {
                "key_list": [[0, 1]],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 2000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": [
                            "presum_rate",
                            "presummed_waveform",
                            "t_sat_lo",
                            "t_sat_hi",
                        ],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/16, period=waveform.period*16, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            },
                            "t_sat_lo, t_sat_hi": {
                                "function": "saturation",
                                "module": "pygama.dsp.processors",
                                "args": ["waveform", 16, "t_sat_lo", "t_sat_hi"],
                                "unit": "ADC",
                            },
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                        "t_sat_lo": "uint16",
                        "t_sat_hi": "uint16",
                    },
                },
            },
            "spms": {
                "key_list": [[3, 4]],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
            },
        }
    }

    raw_dsp_config = """
    {
        "outputs": ["t_sat_lo", "t_sat_hi"],
        "processors": {
            "t_sat_lo, t_sat_hi": {
                "function": "saturation",
                "module": "pygama.dsp.processors",
                "args": ["waveform", 16, "t_sat_lo", "t_sat_hi"],
                "unit": "ADC"
                }
        }
    }
    """
    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)
    lh5_buffer_processor(
        lh5_raw_file_in=raw_file, overwrite=True, out_spec=proc_out_spec
    )

    lh5_tables = lgdo.ls(raw_file)
    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for i, tb in enumerate(lh5_tables):
        if "raw" not in tb and lgdo.ls(raw_file, f"{tb}/raw"):
            lh5_tables[i] = f"{tb}/raw"
        elif not lgdo.ls(raw_file, tb):
            del lh5_tables[i]

    # Grab the proc_spec after keylist expansion from build_raw
    proc_spec = {}
    for key in proc_out_spec["FCEventDecoder"].keys():
        if "proc_spec" in proc_out_spec["FCEventDecoder"][key].keys():
            proc_spec[key] = proc_out_spec["FCEventDecoder"][key].pop("proc_spec")

    jsonfile = proc_spec

    sto = lgdo.LH5Store()

    for raw_group in lh5_tables:
        # First, check the packet ids
        raw_packet_ids, _ = sto.read_object(str(raw_group) + "/packet_id", raw_file)
        processed_packet_ids, _ = sto.read_object(
            str(raw_group) + "/packet_id", processed_file
        )

        assert np.array_equal(raw_packet_ids.nda, processed_packet_ids.nda)

        # Read in the presummed rate from the config file to modify the clock rate later
        group_name = raw_group.split("/raw")[0]
        pass_flag = False
        # If the user passes processing on a group, then the presum_rate is just 1 and there is no windowing
        # If a group_name is absent from the jsonfile, then that means no processing was performed
        if group_name not in jsonfile.keys():
            presum_rate = 1
            window_start_index = 0
            window_end_index = 0
            pass_flag = True
        else:
            presum_rate_string = jsonfile[group_name]["dsp_config"]["processors"][
                "presum_rate, presummed_waveform"
            ]["args"][3]
            presum_rate_start_idx = presum_rate_string.find("/") + 1
            presum_rate_end_idx = presum_rate_string.find(",")
            presum_rate = int(
                presum_rate_string[presum_rate_start_idx:presum_rate_end_idx]
            )

            # This needs to be overwritten with the correct windowing values set in buffer_processor.py
            window_start_index = int(jsonfile[group_name]["window"][1])
            window_end_index = int(jsonfile[group_name]["window"][2])

        # Read in the waveforms
        raw_packet_waveform_values = sto.read_object(
            str(raw_group) + "/waveform/values", raw_file
        )
        if pass_flag:
            presummed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/waveform/values", processed_file
            )
            windowed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/waveform/values", processed_file
            )
        else:
            presummed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/presummed_waveform/values", processed_file
            )
            windowed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/windowed_waveform/values", processed_file
            )

        # Check that the lengths of the waveforms match what we expect
        assert len(raw_packet_waveform_values[0].nda[0]) == presum_rate * len(
            presummed_packet_waveform_values[0].nda[0]
        )
        assert len(raw_packet_waveform_values[0].nda[0]) == len(
            windowed_packet_waveform_values[0].nda[0]
        ) + np.abs(window_start_index) + np.abs(window_end_index)
        assert isinstance(windowed_packet_waveform_values[0].nda[0][0], np.uint16)

        raw_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/waveform/t0", raw_file
        )
        raw_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/waveform/dt", raw_file
        )

        if pass_flag:
            windowed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/waveform/t0", processed_file
            )
            presummed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/waveform/t0", processed_file
            )
        else:
            windowed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/windowed_waveform/t0", processed_file
            )
            presummed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/presummed_waveform/t0", processed_file
            )

        # Check that the t0s match what we expect, with the correct units
        assert (
            raw_packet_waveform_t0s.nda[0]
            == windowed_packet_waveform_t0s.nda[0]
            - raw_packet_waveform_dts.nda[0] * window_start_index
        )
        assert (
            windowed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )
        assert raw_packet_waveform_t0s.nda[0] == presummed_packet_waveform_t0s.nda[0]
        assert (
            presummed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )

        if pass_flag:
            presummed_packet_waveform_dts, _ = sto.read_object(
                str(raw_group) + "/waveform/dt", processed_file
            )
        else:
            presummed_packet_waveform_dts, _ = sto.read_object(
                str(raw_group) + "/presummed_waveform/dt", processed_file
            )
        # Check that the dts match what we expect, with the correct units
        assert (
            raw_packet_waveform_dts.nda[0]
            == presummed_packet_waveform_dts.nda[0] / presum_rate
        )

        # check that the t_lo_sat and t_sat_hi are correct
        if not pass_flag:
            wf_table, _ = sto.read_object(str(raw_group), raw_file)
            pc, _, wf_out = bpc(wf_table, json.loads(raw_dsp_config))
            pc.execute()
            raw_sat_lo = wf_out["t_sat_lo"]
            raw_sat_hi = wf_out["t_sat_hi"]

            proc_sat_lo, _ = sto.read_object(
                str(raw_group) + "/t_sat_lo", processed_file
            )

            proc_sat_hi, _ = sto.read_object(
                str(raw_group) + "/t_sat_hi", processed_file
            )

            assert np.array_equal(raw_sat_lo.nda, proc_sat_lo.nda)
            assert np.array_equal(raw_sat_hi.nda, proc_sat_hi.nda)
            assert type(proc_sat_lo.nda[0]) == np.uint16


# check that packet indexes match in verification test
def test_lh5_buffer_processor_multiple_keys(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca")
    processed_file = daq_file.replace(
        "L200-comm-20220519-phy-geds.orca",
        "L200-comm-20220519-phy-geds-key-test_proc.lh5",
    )
    raw_file = daq_file.replace(
        "L200-comm-20220519-phy-geds.orca", "L200-comm-20220519-phy-geds-key-test.lh5"
    )

    proc_out_spec = {
        "ORFlashCamADCWaveformDecoder": {
            "ch{key}": {
                "key_list": [1028800, 1028801],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
                "proc_spec": {
                    "window": ["waveform", 2000, -1000, "windowed_waveform"],
                    "dsp_config": {
                        "outputs": [
                            "presum_rate",
                            "presummed_waveform",
                            "t_sat_lo",
                            "t_sat_hi",
                        ],
                        "processors": {
                            "presum_rate, presummed_waveform": {
                                "function": "presum",
                                "module": "pygama.dsp.processors",
                                "args": [
                                    "waveform",
                                    0,
                                    "presum_rate",
                                    "presummed_waveform(shape=len(waveform)/16, period=waveform.period*16, offset=waveform.offset)",
                                ],
                                "unit": "ADC",
                            },
                            "t_sat_lo, t_sat_hi": {
                                "function": "saturation",
                                "module": "pygama.dsp.processors",
                                "args": ["waveform", 16, "t_sat_lo", "t_sat_hi"],
                                "unit": "ADC",
                            },
                        },
                    },
                    "drop": ["waveform"],
                    "dtype_conv": {
                        "presummed_waveform/values": "uint32",
                        "t_sat_lo": "uint16",
                        "t_sat_hi": "uint16",
                    },
                },
            },
            "chan{key}": {
                "key_list": [1028803, 1028804],
                "out_stream": processed_file + ":{name}",
                "out_name": "raw",
            },
        }
    }

    raw_out_spec = {
        "ORFlashCamADCWaveformDecoder": {
            "ch{key}": {
                "key_list": [1028800, 1028801],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
            "chan{key}": {
                "key_list": [1028803, 1028804],
                "out_stream": raw_file + ":{name}",
                "out_name": "raw",
            },
        }
    }

    raw_dsp_config = """
    {
        "outputs": ["t_sat_lo", "t_sat_hi"],
        "processors": {
            "t_sat_lo, t_sat_hi": {
                "function": "saturation",
                "module": "pygama.dsp.processors",
                "args": ["waveform", 16, "t_sat_lo", "t_sat_hi"],
                "unit": "ADC"
                }
        }
    }
    """
    # Build the raw file
    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)

    # Do the data processing on the raw file
    lh5_buffer_processor(
        lh5_raw_file_in=raw_file, overwrite=True, out_spec=proc_out_spec
    )

    # Grab the proc_spec after keylist expansion from build_raw
    proc_spec = {}
    for key in proc_out_spec["ORFlashCamADCWaveformDecoder"].keys():
        if "proc_spec" in proc_out_spec["ORFlashCamADCWaveformDecoder"][key].keys():
            proc_spec[key] = proc_out_spec["ORFlashCamADCWaveformDecoder"][key].pop(
                "proc_spec"
            )

    lh5_tables = lgdo.ls(raw_file)
    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for i, tb in enumerate(lh5_tables):
        if "raw" not in tb and lgdo.ls(raw_file, f"{tb}/raw"):
            lh5_tables[i] = f"{tb}/raw"
        elif not lgdo.ls(raw_file, tb):
            del lh5_tables[i]

    jsonfile = proc_spec

    sto = lgdo.LH5Store()

    for raw_group in lh5_tables:
        # First, check the packet ids
        raw_packet_ids, _ = sto.read_object(str(raw_group) + "/packet_id", raw_file)
        processed_packet_ids, _ = sto.read_object(
            str(raw_group) + "/packet_id", processed_file
        )

        assert np.array_equal(raw_packet_ids.nda, processed_packet_ids.nda)

        # Read in the presummed rate from the config file to modify the clock rate later
        group_name = raw_group.split("/raw")[0]

        pass_flag = False
        # If the user passes processing on a group, then the presum_rate is just 1 and there is no windowing
        # If the group_name is absent from the jsonfile, then no processing was done
        if group_name not in jsonfile.keys():
            presum_rate = 1
            window_start_index = 0
            window_end_index = 0
            pass_flag = True
        else:
            presum_rate_string = jsonfile[group_name]["dsp_config"]["processors"][
                "presum_rate, presummed_waveform"
            ]["args"][3]
            presum_rate_start_idx = presum_rate_string.find("/") + 1
            presum_rate_end_idx = presum_rate_string.find(",")
            presum_rate = int(
                presum_rate_string[presum_rate_start_idx:presum_rate_end_idx]
            )

            # This needs to be overwritten with the correct windowing values set in buffer_processor.py
            window_start_index = int(jsonfile[group_name]["window"][1])
            window_end_index = int(jsonfile[group_name]["window"][2])

        # Read in the waveforms
        raw_packet_waveform_values = sto.read_object(
            str(raw_group) + "/waveform/values", raw_file
        )
        if pass_flag:
            presummed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/waveform/values", processed_file
            )
            windowed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/waveform/values", processed_file
            )
        else:
            presummed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/presummed_waveform/values", processed_file
            )
            windowed_packet_waveform_values = sto.read_object(
                str(raw_group) + "/windowed_waveform/values", processed_file
            )

        # Check that the lengths of the waveforms match what we expect
        assert (
            len(raw_packet_waveform_values[0].nda[0])
            // len(presummed_packet_waveform_values[0].nda[0])
            == presum_rate
        )
        assert len(raw_packet_waveform_values[0].nda[0]) == len(
            windowed_packet_waveform_values[0].nda[0]
        ) + np.abs(window_start_index) + np.abs(window_end_index)
        assert isinstance(windowed_packet_waveform_values[0].nda[0][0], np.uint16)

        # Check that the waveforms match
        # These are the channels that should be unprocessed
        if group_name == "chan1028803" or group_name == "chan1028804":
            raw_packet_waveform_values, _ = sto.read_object(
                str(raw_group) + "/waveform/values", raw_file
            )
            windowed_packet_waveform_values, _ = sto.read_object(
                str(raw_group) + "/waveform/values", processed_file
            )
            assert np.array_equal(
                raw_packet_waveform_values.nda, windowed_packet_waveform_values.nda
            )
        else:
            raw_packet_waveform_values, _ = sto.read_object(
                str(raw_group) + "/waveform/values", raw_file
            )
            windowed_packet_waveform_values, _ = sto.read_object(
                str(raw_group) + "/windowed_waveform/values", processed_file
            )
            assert np.array_equal(
                raw_packet_waveform_values.nda[:, window_start_index:window_end_index],
                windowed_packet_waveform_values.nda,
            )

        # Check the t0 and dts are what we expect
        raw_packet_waveform_t0s, _ = sto.read_object(
            str(raw_group) + "/waveform/t0", raw_file
        )
        raw_packet_waveform_dts, _ = sto.read_object(
            str(raw_group) + "/waveform/dt", raw_file
        )

        if pass_flag:
            windowed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/waveform/t0", processed_file
            )
            presummed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/waveform/t0", processed_file
            )
        else:
            windowed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/windowed_waveform/t0", processed_file
            )
            presummed_packet_waveform_t0s, _ = sto.read_object(
                str(raw_group) + "/presummed_waveform/t0", processed_file
            )

        # Check that the t0s match what we expect, with the correct units
        assert (
            raw_packet_waveform_t0s.nda[0]
            == windowed_packet_waveform_t0s.nda[0]
            - raw_packet_waveform_dts.nda[0] * window_start_index
        )
        assert (
            windowed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )
        assert raw_packet_waveform_t0s.nda[0] == presummed_packet_waveform_t0s.nda[0]
        assert (
            presummed_packet_waveform_t0s.attrs["units"]
            == raw_packet_waveform_t0s.attrs["units"]
        )

        if pass_flag:
            presummed_packet_waveform_dts, _ = sto.read_object(
                str(raw_group) + "/waveform/dt", processed_file
            )
        else:
            presummed_packet_waveform_dts, _ = sto.read_object(
                str(raw_group) + "/presummed_waveform/dt", processed_file
            )

            # Check that the presum_rate is correctly identified
            presum_rate_from_file, _ = sto.read_object(
                str(raw_group) + "/presum_rate", processed_file
            )
            assert presum_rate_from_file.nda[0] == presum_rate
        # Check that the dts match what we expect, with the correct units
        assert (
            raw_packet_waveform_dts.nda[0]
            == presummed_packet_waveform_dts.nda[0] / presum_rate
        )

        # check that the t_lo_sat and t_sat_hi are correct
        if not pass_flag:
            wf_table, _ = sto.read_object(str(raw_group), raw_file)
            pc, _, wf_out = bpc(wf_table, json.loads(raw_dsp_config))
            pc.execute()
            raw_sat_lo = wf_out["t_sat_lo"]
            raw_sat_hi = wf_out["t_sat_hi"]

            proc_sat_lo, _ = sto.read_object(
                str(raw_group) + "/t_sat_lo", processed_file
            )

            proc_sat_hi, _ = sto.read_object(
                str(raw_group) + "/t_sat_hi", processed_file
            )

            assert np.array_equal(raw_sat_lo.nda, proc_sat_lo.nda)
            assert np.array_equal(raw_sat_hi.nda, proc_sat_hi.nda)
            assert type(proc_sat_lo.nda[0]) == np.uint16


def test_buffer_processor_all_pass(lgnd_test_data):
    # Set up I/O files, including config
    daq_file = lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca")

    raw_file = daq_file.replace(
        "L200-comm-20220519-phy-geds.orca", "L200-comm-20220519-phy-geds-all-pass.lh5"
    )

    processed_file = daq_file.replace(
        "L200-comm-20220519-phy-geds.orca",
        "L200-comm-20220519-phy-geds-all-pass_proc.lh5",
    )

    proc_out_spec = {
        "*": {
            "{name}": {"key_list": ["*"], "out_stream": processed_file, "proc_spec": {}}
        }
    }

    raw_out_spec = {
        "*": {
            "{name}": {
                "key_list": ["*"],
                "out_stream": raw_file,
            }
        }
    }

    build_raw(in_stream=daq_file, out_spec=raw_out_spec, overwrite=True)
    lh5_buffer_processor(
        lh5_raw_file_in=raw_file, overwrite=True, out_spec=proc_out_spec
    )

    # assert filecmp.cmp(raw_file, processed_file, shallow=True)
    sto = lgdo.LH5Store()
    raw_tables = lgdo.ls(raw_file)
    for tb in raw_tables:
        raw, _ = sto.read_object(tb, raw_file)
        proc, _ = sto.read_object(tb, processed_file)
        if isinstance(raw, lgdo.Scalar):
            raw_value = raw.value
            raw_attrs = raw.attrs
            proc_value = proc.value
            proc_attrs = proc.attrs
            assert raw_value == proc_value
            assert raw_attrs == proc_attrs
        else:
            for obj in raw.keys():
                if not isinstance(raw[obj], lgdo.Table):
                    raw_df = raw.get_dataframe([obj])
                    proc_df = proc.get_dataframe([obj])
                else:
                    for sub_obj in raw[obj].keys():
                        raw_df = raw[obj].get_dataframe([str(sub_obj)])
                        proc_df = proc[obj].get_dataframe([str(sub_obj)])

            assert raw_df.equals(proc_df)



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
    "(n)->(n)",
    **nb_kwargs,
)
def log_check(w_in: np.ndarray, w_log: np.ndarray) -> None:
    """
    Calculate the logarithm of the waveform if all its values
    are positive; otherwise, return NaN.

    Parameters
    ----------
    w_in
        the input waveform.
    w_log
        the output waveform with logged values.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_logged": {
            "function": "log_check",
            "module": "pygama.dsp.processors",
            "args": ["wf_blsub[2100:]", "wf_logged"],
            "unit": "ADC"
        }
    """
    w_log[:] = np.nan

    if np.isnan(w_in).any():
        return

    if np.any(w_in <= 0):
        return

    w_log[:] = np.log(w_in[:])



from pathlib import Path

import pytest

from pygama import lgdo
from pygama.dsp import build_dsp
from pygama.lgdo.lh5_store import LH5Store, ls

config_dir = Path(__file__).parent / "configs"


def test_build_dsp_basics(lgnd_test_data, dsp_test_file):
    with pytest.raises(FileExistsError):
        build_dsp(
            lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
            "/tmp/LDQTA_r117_20200110T105115Z_cal_geds_dsp.lh5",
            dsp_config=f"{config_dir}/icpc-dsp-config.json",
        )

    with pytest.raises(FileNotFoundError):
        build_dsp(
            "non-existent-file.lh5",
            "/tmp/LDQTA_r117_20200110T105115Z_cal_geds_dsp.lh5",
            dsp_config=f"{config_dir}/icpc-dsp-config.json",
            write_mode="r",
        )


def test_build_dsp_spms_channelwise(dsp_test_file_spm):
    assert ls(dsp_test_file_spm) == ["ch0", "ch1", "ch2", "dsp_info"]
    assert ls(dsp_test_file_spm, "ch0/") == ["ch0/dsp"]
    assert ls(dsp_test_file_spm, "ch0/dsp/") == [
        "ch0/dsp/energies",
        "ch0/dsp/trigger_pos",
    ]

    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/ch0/dsp/energies", dsp_test_file_spm)
    assert isinstance(lh5_obj, lgdo.ArrayOfEqualSizedArrays)
    assert len(lh5_obj) == 5
    assert len(lh5_obj.nda[0]) == 20



# Copyright (C) 2003-2005 Peter J. Verveer
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. The name of the author may not be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# All this code belongs to the team that coded Scipy, found at this link:
#     https://github.com/scipy/scipy/blob/v1.6.0/scipy/ndimage/filters.py#L210-L260
# The only thing changed was the calculation of the convulution, which
# originally called a function from a C library.  In this code, the convolution is
# performed with NumPy's built in convolution function.
from __future__ import annotations

import numpy
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def gaussian_filter1d(sigma: int, truncate: float = 4.0) -> numpy.ndarray:
    """1-D Gaussian filter.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Parameters
    ----------
    sigma
        standard deviation for Gaussian kernel
    truncate
        truncate the filter at this many standard deviations.
    """

    def _gaussian_kernel1d(sigma, radius):
        """
        Computes a 1-D Gaussian convolution kernel.
        """
        sigma2 = sigma * sigma
        x = numpy.arange(-radius, radius + 1)
        phi_x = numpy.exp(-0.5 / sigma2 * x**2)
        phi_x = phi_x / phi_x.sum()
        return phi_x

    sd = float(sigma)

    # Make the radius of the filter equal to truncate standard deviations

    lw = int(truncate * sd + 0.5)

    # Since we are calling correlate, not convolve, revert the kernel

    weights = _gaussian_kernel1d(sigma, lw)[::-1]
    weights = numpy.asarray(weights, dtype=numpy.float64)

    # Find the length of the kernel so we can reflect the signal an appropriate amount

    extension_length = int(len(weights) / 2) + 1

    @guvectorize(
        [
            "void(float32[:], float32[:])",
            "void(float64[:], float64[:])",
            "void(int32[:], int32[:])",
            "void(int64[:], int64[:])",
        ],
        "(n),(m)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def gaussian_filter1d_out(wf_in, wf_out):
        # Have to create an array to enable the reflect mode
        # Extend the signal on the left and right by at least half of the length of the kernel

        wf_in = numpy.asarray(wf_in)

        # Short warning message if kernel is larger than signal, in which case signal can't be convolved

        if len(wf_in) < extension_length:
            raise ValueError(
                "Kernel calculated was larger than signal, try again with smaller parameters"
            )

        # This mode extends as a reflection
        # ‘reflect’ (d c b a | a b c d | d c b a)
        # The input is extended by reflecting about the edge of the last pixel. This mode is also
        # sometimes referred to as half-sample symmetric.

        reflected_front = numpy.flip(wf_in[0:extension_length])
        reflected_end = numpy.flip(wf_in[-extension_length:])

        # Extend the signal

        extended_signal = wf_in
        extended_signal = numpy.concatenate((extended_signal, reflected_end), axis=None)
        extended_signal = numpy.concatenate(
            (reflected_front, extended_signal), axis=None
        )
        output = numpy.correlate(extended_signal, weights, mode="same")

        # Now extract the original signal length

        wf_out[:] = output[extension_length:-extension_length]

    return gaussian_filter1d_out



from __future__ import annotations

import numpy as np
from numba import from_dtype, guvectorize, types

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def param_lookup(
    param_dict: dict[int, float], default_val: float, dtype: str | np.dtype
) -> np.ufunc:
    """Generate the :class:`numpy.ufunc` ``lookup(channel, val)``, which
    returns a NumPy array of values corresponding to various channels that are
    looked up in the provided `param_dict`.  If there is no key, use
    `default_val` instead.
    """
    out_type = from_dtype(np.dtype(dtype))

    # convert types to avoid any necessity of casting
    param_dict = {types.uint32(k): out_type(v) for k, v in param_dict.items()}
    default_val = out_type(default_val)

    @guvectorize(
        ["void(uint32, " + out_type.name + "[:])"],
        "()->()",
        **nb_kwargs(cache=False, forceobj=True),
    )
    def lookup(channel: int, val: np.ndarray) -> None:
        """Look up a value for the provided channel from a dictionary provided
        at compile time.
        """
        val[0] = param_dict.get(channel, default_val)

    return lookup



from __future__ import annotations

from pygama.raw.data_decoder import DataDecoder
from pygama.raw.orca.orca_header import OrcaHeader


class OrcaDecoder(DataDecoder):
    """Base class for ORCA decoders.

    Mostly here to provide a standard interface for setting the header during
    initialization.
    """

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.header = None
        if header:
            self.set_header(header)  # allow for derived class function to be called

    def set_header(self, header: OrcaHeader) -> None:
        """Setter for headers. Overload to set card parameters, etc."""
        self.header = header


def get_ccc(crate: int, card: int, channel: int) -> int:
    """Define a standard hash for (crate, card, channel) to integer."""
    return (crate << 9) + ((card & 0x1F) << 4) + (channel & 0xF)


def get_crate(ccc: int) -> int:
    """Get crate number back from hash value.

    See Also
    --------
    .get_ccc
    """
    return ccc >> 9


def get_card(ccc: int) -> int:
    """Get card number back from hash value.

    See Also
    --------
    .get_ccc
    """
    return (ccc >> 4) & 0x1F


def get_channel(ccc: int) -> int:
    """Get channel number back from hash value.

    See Also
    --------
    .get_ccc
    """
    return ccc & 0xF



"""
This module is for extracting a single pole zero constant from the decay tail
"""

from __future__ import annotations

import json
import logging
import os
import pathlib
import pickle as pkl
from collections import OrderedDict

import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt
import numpy as np

import pygama
import pygama.lgdo as lgdo
import pygama.lgdo.lh5_store as lh5
import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.pargen.cuts as cts
import pygama.pargen.dsp_optimize as opt
import pygama.pargen.energy_optimisation as om

log = logging.getLogger(__name__)


def load_data(
    raw_file: list[str],
    lh5_path: str,
    n_events: int = 10000,
    threshold: int = 5000,
    wf_field: str = "waveform",
) -> lgdo.Table:
    sto = lh5.LH5Store()
    df = lh5.load_dfs(raw_file, ["daqenergy", "timestamp"], lh5_path)

    pulser_props = cts.find_pulser_properties(df, energy="daqenergy")
    if len(pulser_props) > 0:
        final_mask = None
        for entry in pulser_props:
            e_cut = (df.daqenergy.values < entry[0] + entry[1]) & (
                df.daqenergy.values > entry[0] - entry[1]
            )
            if final_mask is None:
                final_mask = e_cut
            else:
                final_mask = final_mask | e_cut
        ids = ~(final_mask)
        log.debug(f"pulser found: {pulser_props}")
    else:
        log.debug("no_pulser")
        ids = np.ones(len(df.daqenergy.values), dtype=bool)

    cuts = np.where((df.daqenergy.values > threshold) & (ids))[0]

    waveforms = sto.read_object(
        f"{lh5_path}/{wf_field}", raw_file, idx=cuts, n_rows=n_events
    )[0]
    baseline = sto.read_object(
        f"{lh5_path}/baseline", raw_file, idx=cuts, n_rows=n_events
    )[0]
    tb_data = lh5.Table(col_dict={f"{wf_field}": waveforms, "baseline": baseline})
    return tb_data


def get_decay_constant(
    slopes: np.array, wfs: lgdo.WaveformTable, display: int = 0
) -> dict:
    """
    Finds the decay constant from the modal value of the tail slope after cuts
    and saves it to the specified json.

    Parameters
    ----------
    slopes : array
        tail slope array

    dict_file : str
        path to json file to save decay constant value to.
        It will be saved as a dictionary of form {'pz': {'tau': decay_constant}}

    Returns
    -------
    tau_dict : dict
    """
    tau_dict = {}

    pz = tau_dict.get("pz")

    counts, bins, var = pgh.get_hist(slopes, bins=100000, range=(-0.01, 0))
    bin_centres = pgh.get_bin_centers(bins)
    high_bin = bin_centres[np.argmax(counts)]
    try:
        pars, cov = pgf.gauss_mode_width_max(
            counts,
            bins,
            n_bins=10,
            cost_func="Least Squares",
            inflate_errors=False,
            gof_method="var",
        )
        if np.abs(np.abs(pars[0] - high_bin) / high_bin) > 0.05:
            raise ValueError
        high_bin = pars[0]
    except:
        pass
    tau = round(-1 / (high_bin), 1)

    sampling_rate = wfs["dt"].nda[0]
    units = wfs["dt"].attrs["units"]
    tau = f"{tau*sampling_rate}*{units}"

    tau_dict["pz"] = {"tau": tau}
    if display > 0:
        out_plot_dict = {}
        plt.rcParams["figure.figsize"] = (10, 6)
        plt.rcParams["font.size"] = 8
        fig, ax = plt.subplots()
        bins = np.linspace(-0.01, 0, 100000)  # change if needed
        counts, bins, bars = ax.hist(slopes, bins=bins, histtype="step")
        plot_max = np.argmax(counts)
        in_min = plot_max - 20
        if in_min < 0:
            in_min = 0
        in_max = plot_max + 21
        if in_max >= len(bins):
            in_min = len(bins) - 1
        plt.xlabel("Slope")
        plt.ylabel("Counts")
        plt.yscale("log")
        axins = ax.inset_axes([0.5, 0.45, 0.47, 0.47])
        axins.hist(
            slopes[(slopes > bins[in_min]) & (slopes < bins[in_max])],
            bins=200,
            histtype="step",
        )
        axins.axvline(high_bin, color="red")
        axins.set_xlim(bins[in_min], bins[in_max])
        labels = ax.get_xticklabels()
        ax.set_xticklabels(labels=labels, rotation=45)
        out_plot_dict["slope"] = fig
        if display > 1:
            plt.show()
        else:
            plt.close()
        return tau_dict, out_plot_dict
    else:
        return tau_dict


def fom_dpz(tb_data, verbosity=0, rand_arg=None):
    std = tb_data["pz_std"].nda
    counts, start_bins, var = pgh.get_hist(std, dx=0.1, range=(0, 400))
    max_idx = np.argmax(counts)
    mu = start_bins[max_idx]
    try:
        pars, cov = pgf.gauss_mode_width_max(
            counts,
            start_bins,
            mode_guess=mu,
            n_bins=10,
            cost_func="Least Squares",
            inflate_errors=False,
            gof_method="var",
        )

        mu = pars[0]

    except:
        mu = start_bins[max_idx]

    return {"y_val": np.abs(mu)}


def get_dpz_consts(grid_out, opt_dict):
    std_grid = np.ndarray(shape=grid_out.shape)
    for i in range(grid_out.shape[0]):
        for j in range(grid_out.shape[1]):
            std_grid[i, j] = grid_out[i, j]["y_val"]
    min_val = np.amin(std_grid)
    min_point = np.where(std_grid == np.amin(std_grid))

    opt_name = list(opt_dict.keys())[0]
    keys = list(opt_dict[opt_name].keys())
    param_list = []
    shape = []
    db_dict = {}
    for key in keys:
        param_dict = opt_dict[opt_name][key]
        grid_axis = np.arange(
            param_dict["start"], param_dict["end"], param_dict["spacing"]
        )
        unit = param_dict.get("unit")
        param_list.append(grid_axis)
        shape.append(len(grid_axis))
    for i, key in enumerate(keys):
        unit = opt_dict[opt_name][key].get("unit")

        if unit is not None:
            try:
                db_dict[opt_name].update(
                    {key: f"{param_list[i][min_point[i]][0]}*{unit}"}
                )
            except:
                db_dict[opt_name] = {key: f"{param_list[i][min_point[i]][0]}*{unit}"}
        else:
            try:
                db_dict[opt_name].update({key: f"{param_list[i][min_point[i]][0]}"})
            except:
                db_dict[opt_name] = {key: f"{param_list[i][min_point[i]][0]}"}
    return db_dict


def dsp_preprocess_decay_const(
    raw_files: list[str],
    dsp_config: dict,
    lh5_path: str,
    double_pz: bool = False,
    display: int = 0,
    opt_dict: dict = None,
    threshold: int = 5000,
    wf_field: str = "waveform",
    wf_plot: str = "wf_pz",
    norm_param: str = "pz_mean",
    cut_parameters: dict = {"bl_mean": 4, "bl_std": 4, "bl_slope": 4},
) -> dict:
    """
    This function calculates the pole zero constant for the input data

    Parameters
    ----------
    f_raw : str
        The raw file to run the macro on
    dsp_config: str
        Path to the dsp config file, this is a stripped down version which just includes cuts and slope of decay tail
    channel:  str
        Name of channel to process, should be name of lh5 group in raw files

    Returns
    -------
    tau_dict : dict
    """

    tb_data = load_data(raw_files, lh5_path, wf_field=wf_field, threshold=threshold)
    tb_out = opt.run_one_dsp(tb_data, dsp_config)
    log.debug("Processed Data")
    cut_dict = cts.generate_cuts(tb_out, parameters=cut_parameters)
    log.debug("Generated Cuts:", cut_dict)
    idxs = cts.get_cut_indexes(tb_out, cut_dict)
    log.debug("Applied cuts")
    log.debug(f"{len(idxs)} events passed cuts")
    slopes = tb_out["tail_slope"].nda
    log.debug("Calculating pz constant")
    if display > 0:
        tau_dict, plot_dict = get_decay_constant(
            slopes[idxs], tb_data[wf_field], display=display
        )
    else:
        tau_dict = get_decay_constant(slopes[idxs], tb_data[wf_field])
    if double_pz == True:
        log.debug("Calculating double pz constants")
        pspace = om.set_par_space(opt_dict)
        grid_out = opt.run_grid(
            tb_data, dsp_config, pspace, fom_dpz, tau_dict, fom_kwargs=None
        )
        out_dict = get_dpz_consts(grid_out, opt_dict)
        tau_dict["pz"].update(out_dict["pz"])
    if display > 0:
        tb_out = opt.run_one_dsp(tb_data, dsp_config, db_dict=tau_dict)
        wfs = tb_out[wf_plot]["values"].nda[idxs]
        wf_idxs = np.random.choice(len(wfs), 100)
        if norm_param is not None:
            means = tb_out[norm_param].nda[idxs][wf_idxs]
            wfs = np.divide(wfs[wf_idxs], np.reshape(means, (len(wf_idxs), 1)))
        else:
            wfs = wfs[wf_idxs]
        fig2 = plt.figure()
        for wf in wfs:
            plt.plot(np.arange(0, len(wf), 1), wf)
        plt.axhline(1, color="black")
        plt.axhline(0, color="black")
        plt.xlabel("Samples")
        plt.ylabel("ADU")
        plot_dict["waveforms"] = fig2
        if display > 1:
            plt.show()
        else:
            plt.close()
        return tau_dict, plot_dict
    else:
        return tau_dict



import pytest

from pygama.raw.orca.orca_streamer import OrcaStreamer


@pytest.fixture(scope="module")
def orca_stream(lgnd_test_data):
    orstr = OrcaStreamer()
    orstr.open_stream(
        lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca")
    )
    return orstr



from __future__ import annotations

import json
import logging

from pygama import lgdo

log = logging.getLogger(__name__)


class OrcaHeader(dict):
    """ORCA file header object."""

    def __init__(self, jsons: str = None, lgdo_scalar: lgdo.Scalar = None) -> None:
        if jsons is not None:
            self.update(json.loads(jsons))
        elif lgdo_scalar is not None:
            self.set_from_lgdo(lgdo_scalar)

    def set_from_lgdo(self, lgdo_scalar: lgdo.Scalar) -> None:
        if not isinstance(lgdo_scalar, lgdo.Scalar):
            raise ValueError(f"can't instantiate a header from a {type(lgdo_scalar)}")
        else:
            self.update(json.loads(lgdo_scalar.value))

    def get_decoder_list(self) -> list[str]:
        return list(set(self.get_id_to_decoder_name_dict().values()))

    def get_id_to_decoder_name_dict(self, shift_data_id: bool = True) -> dict[int, str]:
        id_dict = {0: "OrcaHeaderDecoder"}
        dd = self["dataDescription"]
        for class_key in dd.keys():
            for super_key in dd[class_key].keys():
                data_id = dd[class_key][super_key]["dataId"]
                if shift_data_id:
                    if data_id < 0:  # short record
                        data_id = (-data_id) >> 26
                    else:
                        data_id = data_id >> 18
                id_dict[data_id] = dd[class_key][super_key]["decoder"]
        return id_dict

    def get_run_number(self) -> int:
        for d in self["ObjectInfo"]["DataChain"]:
            if "Run Control" in d:
                return d["Run Control"]["RunNumber"]
        raise ValueError("No run number found in header!")

    def get_object_info(self, orca_class_name: str) -> dict[int, dict[int, dict]]:
        """Returns a ``dict[crate][card]`` with all info from the header for
        each card with name `orca_class_name`.
        """
        object_info_dict = {}

        crates = self["ObjectInfo"]["Crates"]
        for crate in crates:
            for card in crate["Cards"]:
                if card["Class Name"] == orca_class_name:
                    if crate["CrateNumber"] not in object_info_dict:
                        object_info_dict[crate["CrateNumber"]] = {}
                    object_info_dict[crate["CrateNumber"]][card["Card"]] = card

        return object_info_dict

    def get_readout_info(self, orca_class_name: str, unique_id: int = -1) -> list:
        """Returns a list with all the readout list info from the header with
        name `orca_class_name`.

        Optionally, if `unique_id` is greater or equal than zero, only return
        the readout info for that ORCA unique ID number.
        """
        readout_info_list = []
        try:
            readouts = self["ReadoutDescription"]
            for readout in readouts:
                try:
                    if readout["name"] == orca_class_name:
                        if unique_id >= 0 and readout["uniqueID"] == unique_id:
                            return readout
                        readout_info_list.append(readout)
                except KeyError:
                    continue
        except KeyError:
            pass
        if len(readout_info_list) == 0:
            log.warning(f"no readout info for class name '{orca_class_name}'")
        return readout_info_list

    def get_auxhw_info(self, orca_class_name: str, unique_id: int = -1) -> list:
        """Returns a list with all the info from the ``AuxHw`` table of the header
        with name `orca_class_name`.

        Optionally, if `unique_id` is greater or equal to zero, only return the
        object for that ORCA unique ID number.
        """
        auxhw_info_list = []
        try:
            objs = self["ObjectInfo"]["AuxHw"]
            for obj in objs:
                try:
                    if obj["Class Name"] == orca_class_name:
                        if unique_id >= 0 and obj["uniqueID"] == unique_id:
                            return obj
                        auxhw_info_list.append(obj)
                except KeyError:
                    continue
        except KeyError:
            pass
        if len(auxhw_info_list) == 0:
            log.warning(f"no object info for class name '{orca_class_name}'")
        return auxhw_info_list



from pygama import lgdo


def test_decoding(fcio_config):
    pass


def test_data_types(fcio_config):
    assert isinstance(fcio_config, lgdo.Struct)


def test_values(fcio_config):
    expected_dict = {
        "nsamples": 6000,
        "nadcs": 6,
        "ntriggers": 0,
        "telid": 0,
        "adcbits": 16,
        "sumlength": 1,
        "blprecision": 1,
        "mastercards": 1,
        "triggercards": 0,
        "adccards": 1,
        "gps": 0,
    }

    for k, v in expected_dict.items():
        assert fcio_config[k].value == v



import numpy as np

from pygama import lgdo
from pygama.lgdo import Array


def test_datatype_name():
    array = Array()
    assert array.datatype_name() == "array"


def test_form_datatype():
    array = Array(shape=(12, 34))
    assert array.form_datatype() == "array<2>{real}"


def test_init():
    attrs = {"attr1": 1}
    array = Array(shape=(3,), dtype=np.float32, fill_val=42, attrs=attrs)
    assert (array.nda == np.full((3,), 42, np.float32)).all()
    assert array.attrs == attrs | {"datatype": "array<1>{real}"}


def test_resize():
    array = Array(nda=np.array([1, 2, 3, 4]))
    array.resize(3)
    assert (array.nda == np.array([1, 2, 3])).all()


def test_copy():
    a1 = Array(np.array([1, 2, 3, 4]))
    a2 = lgdo.copy(a1)
    assert a1 == a2


def test_insert():
    a = Array(np.array([1, 2, 3, 4]))
    a.insert(2, [-1, -1])
    assert a == Array([1, 2, -1, -1, 3, 4])



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs

from .time_point_thresh import time_point_thresh


@guvectorize(
    [
        "void(float32[:],float32[:],float32[:])",
        "void(float64[:],float64[:],float64[:])",
    ],
    "(n),(n) -> (n)",
    **nb_kwargs,
    forceobj=True,
)
def remove_duplicates(
    t_in: np.ndarray, vt_min_in: np.ndarray, t_out: np.ndarray
) -> None:
    """Helper function to remove duplicate peak positions.

    :func:`.time_point_thresh` has issues with afterpulsing in waveforms that
    causes an aferpulse peak position to be sent to zero or the same index as
    the first pulse. This only happens when the relative minimum between the
    first pulse and the afterpulse is greater than the threshold. So, we sweep
    through the array again to ensure there are no duplicate indices. If there
    are duplicate indices caused by a misidentified position of an afterpulse,
    we replace its index by that of the corresponding minimum found using
    :func:`.get_multi_local_extrema`. It also checks to make sure that the
    maximum of a waveform isn't right at index 0.

    Parameters
    ----------
    t_in
        the array of indices that we want to remove duplicates from.
    vt_min_in
        list of indices of minima that we want to replace duplicates in `t_out`
        with.
    t_out
        the array we want to return that will have no duplicate indices in it.

    See Also
    --------
    .multi_t_filter
    """
    # initialize arrays
    t_out[:] = np.nan

    # checks
    if (
        np.isnan(t_in).all() and np.isnan(vt_min_in).all()
    ):  # we pad these with NaNs, so only return if there is nothing to analyze
        return

    # check if any later indexed values are equal to the earliest instance
    k = 0
    for index1 in range(len(t_in)):
        for index2 in range(len(t_in[index1 + 1 :])):
            if t_in[index1] == t_in[index2 + index1 + 1]:
                t_out[index2 + index1 + 1] = vt_min_in[k]
        k += 1  # this makes sure that the index of the misidentified afterpulse tp0 is replaced with the correct corresponding minimum

    # Fill up the output with the rest of the values from the input that weren't repeats
    for index in range(len(t_in)):
        if np.isnan(t_out[index]) and not np.isnan(t_in[index]):
            t_out[index] = t_in[index]

    # makes sure that the first maximum found isn't the start of the waveform
    if not np.isnan(t_out[0]):
        if int(t_out[0]) == 0:
            t_out[:] = np.append(t_out[1:], np.nan)


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:], float64[:], float64[:])",
    ],
    "(n),(),(m),(m),(m)",
    **nb_kwargs,
    forceobj=True,
)
def multi_t_filter(
    w_in: np.ndarray,
    a_threshold_in: float,
    vt_max_in: np.ndarray,
    vt_min_in: np.ndarray,
    t_out: np.ndarray,
) -> None:
    """Gets list of indices of the start of leading edges of multiple peaks
    within a waveform.

    Is built to handle afterpulses/delayed cross-talk and trains of pulses.
    Works by calling the vectorized functions :func:`.get_multi_local_extrema`
    which returns a list of the maxima and minima in a waveform, and then the
    list of maxima is fed into :func:`.time_point_thresh` which returns the
    final times that waveform is less than a specified threshold.

    Parameters
    ----------
    w_in
        the array of data within which the list of leading edge times will be
        found.
    a_threshold_in
        threshold to search for using :func:`.time_point_thresh`.
    vt_max_in
        the array of maximum positions for each waveform.
    vt_min_in
        the array of minimum positions for each waveform.
    t_out
        output array of fixed length (padded with :any:`numpy.nan`) that hold
        the indices of the identified initial rise times of peaks in the
        signal.

    See Also
    --------
    ~.time_point_thresh.time_point_thresh
    """

    # initialize arrays, padded with the elements we want
    t_out[:] = np.nan

    # checks
    if np.isnan(w_in).any() or np.isnan(a_threshold_in):
        return
    if np.isnan(vt_max_in).all() and np.isnan(vt_min_in).all():
        return
    if not len(t_out) <= len(w_in):
        raise DSPFatal(
            "The length of your return array must be smaller than the length of your waveform"
        )

    # Initialize an intermediate array to hold the tp0 values before we remove duplicates from it
    intermediate_t_out = np.full_like(t_out, np.nan, dtype=np.float32)

    # Go through the list of maxima, calling time_point_thresh (the refactored version ignores the nan padding)
    time_point_thresh(w_in, a_threshold_in, vt_max_in, 0, intermediate_t_out)

    # Remove duplicates from the t_out list
    remove_duplicates(intermediate_t_out, vt_min_in, t_out)



"""
Provides convenience functions for working with ORCA packets.

An ORCA packet is represented by a one-dimensional :class:`numpy.ndarray` of
type :class:`numpy.uint32`.
"""
from __future__ import annotations

import logging

import numpy as np
import numpy.typing as npt

log = logging.getLogger(__name__)
OrcaPacket = npt.NDArray[np.uint32]


def is_short(packet: OrcaPacket) -> bool:
    return bool(packet[0] >> 31)


def get_n_words(packet: OrcaPacket) -> int:
    if is_short(packet):
        return 1
    return packet[0] & 0x3FFFF


def get_data_id(packet: OrcaPacket, shift: bool = True) -> np.uint32:
    # in the header the data_id appears unshifted, but in pygama we use it
    # shifted by default so that it has sensible values when printed / checked
    if is_short(packet):
        if shift:
            return (packet[0] & 0xFC000000) >> 26
        else:
            return packet[0] & 0xFC000000
    if shift:
        return (packet[0] & 0xFFFC0000) >> 18
    return packet[0] & 0xFFFC0000


def hex_dump(
    packet: OrcaPacket,
    shift_data_id: bool = True,
    print_n_words: bool = False,
    max_words: int = np.inf,
    as_int: bool = False,
    as_short: bool = False,
    id_dict: dict = None,
    use_logging: bool = True,
) -> None:
    dump_cmd = print  # noqa: T202
    if use_logging:
        dump_cmd = log.debug

    data_id = get_data_id(packet, shift=shift_data_id)
    n_words = get_n_words(packet)
    if id_dict is not None:
        if data_id in id_dict:
            heading = f"{id_dict[data_id]} (data ID = {data_id})"
        else:
            heading = f"[unknown] (data ID = {data_id})"
    else:
        heading = f"data ID = {data_id}"
    if print_n_words:
        dump_cmd(f"{heading}: {n_words} words")
    else:
        dump_cmd(f"{heading}:")
        n_to_print = int(np.minimum(n_words, max_words))
        pad = int(np.ceil(np.log10(n_to_print)))
        for i in range(n_to_print):
            line = f"{str(i).zfill(pad)}"
            line += " {0:#0{1}x}".format(packet[i], 10)
            if data_id == 0 and i > 1:
                line += f" {packet[i:i+1].tobytes().decode()}"
            if as_int:
                line += f" {packet[i]}"
            if as_short:
                line += f" {np.frombuffer(packet[i:i+1].tobytes(), dtype='uint16')}"
            dump_cmd(line)



"""
Pygama: decoding and processing digitizer data.
Check out the `online documentation <https://pygama.readthedocs.io>`_
"""

from pygama._version import version as __version__

__all__ = ["__version__"]



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32[:], float32, float32, float32[:], float32[:])",
        "void(float64[:], float64[:], float64, float64, float64[:], float64[:])",
    ],
    "(n),(m),(),(),(m),()",
    **nb_kwargs,
)
def peak_snr_threshold(
    w_in: np.ndarray,
    idx_in: np.ndarray,
    ratio_in: float,
    width_in: int,
    idx_out: np.ndarray,
    n_idx_out: int,
) -> None:
    """Search for local minima in a window around the provided indices.

    If a minimum is found it is checked whether amplitude of the minimum
    divided by the amplitude of the waveform at the index is smaller then the
    given ratio. If this is the case the index is copied to the output.

    Parameters
    ----------
    w_in
        the input waveform.
    idx_in
        the array of indices of possible signal candidates.
    ratio_in
        ratio threshold value.
    width_in
        width of the local search window.
    idx_out
        indices of local minima.
    n_idx_out
        number of non-:any:`numpy.nan` indices in `idx_out`.
    """

    # prepare output

    idx_out[:] = np.nan
    n_idx_out[0] = 0

    # fill output
    k = 0
    for i in range(len(idx_in)):
        if not np.isnan(idx_in[i]):
            a = int(idx_in[i]) - int(width_in)
            b = int(idx_in[i]) + int(width_in)
            if a < 0:
                a = 0
            if b >= len(w_in):
                b = len(w_in) - 1
            min_index = a
            for j in range(a, b, 1):
                if w_in[j] < w_in[min_index]:
                    min_index = j
            if np.absolute(w_in[min_index] / w_in[int(idx_in[i])]) < ratio_in:
                idx_out[k] = idx_in[i]
                k = k + 1
    n_idx_out[0] = k



from __future__ import annotations

from typing import Callable

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def cusp_filter(length: int, sigma: float, flat: int, decay: int) -> Callable:
    """Apply a CUSP filter to the waveform.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Parameters
    ----------
    length
        the length of the filter to be convolved.
    sigma
        the curvature of the rising and falling part of the kernel.
    flat
        the length of the flat section.
    decay
        the decay constant of the exponential to be convolved.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_cusp": {
            "function": "cusp_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_bl", "wf_cusp(101,f)"],
            "unit": "ADC",
            "init_args": ["len(wf_bl)-100", "40*us", "3*us", "45*us"]
        }
    """
    if length <= 0:
        raise DSPFatal("The length of the filter must be positive")

    if np.floor(length) != length:
        raise DSPFatal("The length of the filter must be an integer")

    if sigma < 0:
        raise DSPFatal("The curvature parameter must be positive")

    if flat < 0:
        raise DSPFatal("The length of the flat section must be positive")

    if np.floor(flat) != flat:
        raise DSPFatal("The length of the flat section must be an integer")

    if decay < 0:
        raise DSPFatal("The decay constant must be positive")

    lt = int((length - flat) / 2)
    flat_int = int(flat)
    cusp = np.zeros(length)
    for ind in range(0, lt, 1):
        cusp[ind] = float(np.sinh(ind / sigma) / np.sinh(lt / sigma))
    for ind in range(lt, lt + flat_int + 1, 1):
        cusp[ind] = 1
    for ind in range(lt + flat_int + 1, length, 1):
        cusp[ind] = float(np.sinh((length - ind) / sigma) / np.sinh(lt / sigma))

    den = [1, -np.exp(-1 / decay)]
    cuspd = np.convolve(cusp, den, "same")

    @guvectorize(
        ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
        "(n),(m)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def cusp_out(w_in: np.ndarray, w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        w_in
            the input waveform.
        w_out
            the filtered waveform.
        """
        w_out[:] = np.nan

        if np.isnan(w_in).any():
            return

        if len(cuspd) > len(w_in):
            raise DSPFatal("The filter is longer than the input waveform")

        w_out[:] = np.convolve(w_in, cuspd, "valid")

    return cusp_out


def zac_filter(length: int, sigma: float, flat: int, decay: int) -> Callable:
    """Apply a ZAC (Zero Area CUSP) filter to the waveform.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Parameters
    ----------
    length
        the length of the filter to be convolved.
    sigma
        the curvature of the rising and falling part of the kernel.
    flat
        the length of the flat section.
    decay
        the decay constant of the exponential to be convolved.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_zac": {
            "function": "zac_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_bl", "wf_zac(101,f)"],
            "unit": "ADC",
            "init_args": ["len(wf_bl)-100", "40*us", "3*us", "45*us"],
        }
    """
    if length <= 0:
        raise DSPFatal("The length of the filter must be positive")

    if np.floor(length) != length:
        raise DSPFatal("The length of the filter must be an integer")

    if sigma < 0:
        raise DSPFatal("The curvature parameter must be positive")

    if flat < 0:
        raise DSPFatal("The length of the flat section must be positive")

    if np.floor(flat) != flat:
        raise DSPFatal("The length of the flat section must be an integer")

    if decay < 0:
        raise DSPFatal("The decay constant must be positive")

    lt = int((length - flat) / 2)
    flat_int = int(flat)

    # calculate cusp filter and negative parables
    cusp = np.zeros(length)
    par = np.zeros(length)
    for ind in range(0, lt, 1):
        cusp[ind] = float(np.sinh(ind / sigma) / np.sinh(lt / sigma))
        par[ind] = np.power(ind - lt / 2, 2) - np.power(lt / 2, 2)
    for ind in range(lt, lt + flat_int + 1, 1):
        cusp[ind] = 1
    for ind in range(lt + flat_int + 1, length, 1):
        cusp[ind] = float(np.sinh((length - ind) / sigma) / np.sinh(lt / sigma))
        par[ind] = np.power(length - ind - lt / 2, 2) - np.power(lt / 2, 2)

    # calculate area of cusp and parables
    areapar, areacusp = 0, 0
    for i in range(0, length, 1):
        areapar += par[i]
        areacusp += cusp[i]

    # normalize parables area
    par = -par / areapar * areacusp

    # create zac filter
    zac = cusp + par

    # deconvolve zac filter
    den = [1, -np.exp(-1 / decay)]
    zacd = np.convolve(zac, den, "same")

    @guvectorize(
        ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
        "(n),(m)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def zac_out(w_in: np.ndarray, w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        w_in
            the input waveform.
        w_out
            the filtered waveform.
        """
        w_out[:] = np.nan

        if np.isnan(w_in).any():
            return

        if len(zacd) > len(w_in):
            raise DSPFatal("The filter is longer than the input waveform")

        w_out[:] = np.convolve(w_in, zacd, "valid")

    return zac_out


def t0_filter(rise: int, fall: int) -> Callable:
    """Apply a modified, asymmetric trapezoidal filter to the waveform.

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Parameters
    ----------
    rise
        the length of the rise section. This is the linearly increasing
        section of the filter that performs a weighted average.
    fall
        the length of the fall section. This is the simple averaging part
        of the filter.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_t0_filter": {
            "function": "t0_filter",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "wf_t0_filter(3748,f)"],
            "unit": "ADC",
            "init_args": ["128*ns", "2*us"]
        }
    """
    if rise < 0:
        raise DSPFatal("The length of the rise section must be positive")

    if fall < 0:
        raise DSPFatal("The length of the fall section must be positive")

    t0_kern = np.zeros(int(rise) + int(fall))
    for i in range(int(rise)):
        t0_kern[i] = 2 * (int(rise) - i) / (rise**2)
    for i in range(int(rise), len(t0_kern), 1):
        t0_kern[i] = -1 / fall

    @guvectorize(
        ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
        "(n),(m)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def t0_filter_out(w_in: np.ndarray, w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        w_in
            The input waveform
        w_out
            The filtered waveform
        """
        w_out[:] = np.nan

        if np.isnan(w_in).any():
            return

        if len(t0_kern) > len(w_in):
            raise DSPFatal("The filter is longer than the input waveform")

        w_out[:] = np.convolve(w_in, t0_kern)[: len(w_in)]

    return t0_filter_out



import pytest

from pygama import lgdo
from pygama.raw.compass.compass_event_decoder import CompassEventDecoder
from pygama.raw.raw_buffer import RawBuffer


@pytest.fixture(scope="module")
def event_rbkd(lgnd_test_data, compass_config, packet):
    decoder = CompassEventDecoder()
    decoder.set_header(compass_config)

    # build raw buffer for each channel in the FC trace list
    rbkd = {}
    rbkd[0] = RawBuffer(lgdo=decoder.make_lgdo(size=1))

    # decode packet into the lgdo's and check if the buffer is full
    assert (
        decoder.decode_packet(packet, evt_rbkd=rbkd, packet_id=1, header=compass_config)
        is True
    )
    return rbkd


def test_decoding(event_rbkd):
    assert event_rbkd != {}


def test_data_types(event_rbkd):
    for _, v in event_rbkd.items():
        tbl = v.lgdo
        assert isinstance(tbl, lgdo.Struct)
        assert isinstance(tbl["packet_id"], lgdo.Array)
        assert isinstance(tbl["board"], lgdo.Array)
        assert isinstance(tbl["channel"], lgdo.Array)
        assert isinstance(tbl["timestamp"], lgdo.Array)
        assert isinstance(tbl["energy"], lgdo.Array)
        assert isinstance(tbl["energy_short"], lgdo.Array)
        assert isinstance(tbl["flags"], lgdo.Array)
        assert isinstance(tbl["num_samples"], lgdo.Array)
        assert isinstance(tbl["waveform"], lgdo.Struct)
        assert isinstance(tbl["waveform"]["t0"], lgdo.Array)
        assert isinstance(tbl["waveform"]["dt"], lgdo.Array)
        assert isinstance(tbl["waveform"]["values"], lgdo.ArrayOfEqualSizedArrays)


def test_values(event_rbkd):
    for _, v in event_rbkd.items():
        tbl = v.lgdo
        assert tbl["packet_id"].nda == [1]
        assert tbl["board"].nda == 0
        assert tbl["channel"].nda == 0
        assert tbl["timestamp"].nda == [9.78762e10]
        assert tbl["energy"].nda == [798]
        assert tbl["energy_short"].nda == [135]
        assert tbl["flags"].nda == [16384]
        assert tbl["num_samples"].nda == 1000
        assert (
            tbl["waveform"]["values"].nda[0][:14]
            == [
                2745,
                2742,
                2745,
                2746,
                2745,
                2743,
                2745,
                2744,
                2746,
                2747,
                2746,
                2743,
                2745,
                2747,
            ]
        ).all()



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->()",
    **nb_kwargs,
)
def time_over_threshold(w_in: np.ndarray, a_threshold: float, n_samples: float) -> None:
    """Calculates the number of samples in the input waveform over a_threshold

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        the threshold value.
    n_samples
        the number of samples over the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "t_sat": {
            "function": "time_over_threshold",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "a_threshold", "t_sat"],
            "unit": "ns"
        }
    """
    if np.isnan(w_in).any() or np.isnan(a_threshold):
        n_samples[0] = np.nan
        return
    n_samples[0] = 0.0
    for sample in w_in[:]:
        if sample > a_threshold:
            n_samples[0] += 1



from __future__ import annotations

from typing import Callable

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


def dplms(
    length: int,
    noise_mat: list,
    reference: list,
    a1: float,
    a2: float,
    a3: float,
    ff: int,
    coefficients: list,
) -> Callable:
    """Calculate and apply an optimum DPLMS filter to the waveform.

    The processor takes the noise matrix and the reference signal as input and
    calculates the optimum filter according to the provided length and
    penalized coefficients [DPLMS]_.

    .. [DPLMS] V. D'Andrea et al. “Optimum Filter Synthesis with DPLMS
        Method for Energy Reconstruction” Eur. Phys. J. C 83, 149 (2023).
        https://doi.org/10.1140/epjc/s10052-023-11299-z

    Note
    ----
    This processor is composed of a factory function that is called using the
    `init_args` argument. The input and output waveforms are passed using
    `args`.

    Parameters
    ----------
    noise_mat
        noise matrix
    reference
        reference signal
    length
        length of the calculated filter
    a1
        penalized coefficient for the noise matrix
    a2
        penalized coefficient for the reference matrix
    a3
        penalized coefficient for the zero area matrix
    ff
        flat top length for the reference signal
    coefficients
        filter coefficients (if given avoid filter calculation)


    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_dplms": {
            "function": "dplms",
            "module": "pygama.dsp.processors",
            "args": ["wf_diff", "wf_dplms(len(wf_diff)-49, 'f')"],
            "unit": "ADC",
            "init_args": [
                "db.dplms.noise_matrix",
                "db.dplms.reference",
                "50", "0.1", "1", "0", "1"]
        }
    """

    if length <= 0:
        raise DSPFatal("The length of the filter must be positive")

    if len(coefficients) > 0:
        if length != len(coefficients):
            raise DSPFatal("The length and the provided filter are not matching")

        x = np.array(coefficients)
    else:
        noise_mat = np.array(noise_mat)
        reference = np.array(reference)

        if length != noise_mat.shape[0]:
            raise DSPFatal(
                "The length of the filter is not consistent with the noise matrix"
            )

        if len(reference) <= 0:
            raise DSPFatal("The length of the reference signal must be positive")

        if a1 <= 0:
            raise DSPFatal("The penalized coefficient for the noise must be positive")

        if a2 <= 0:
            raise DSPFatal(
                "The penalized coefficient for the reference must be positive"
            )

        if a3 <= 0:
            raise DSPFatal(
                "The penalized coefficient for the zero area must be positive"
            )

        if ff <= 0:
            raise DSPFatal(
                "The penalized coefficient for the ref matrix must be positive"
            )

        # reference matrix
        ssize = len(reference)
        flo = int(ssize / 2 - length / 2)
        fhi = int(ssize / 2 + length / 2)
        ref_mat = np.zeros([length, length])
        ref_sig = np.zeros([length])
        if ff == 0:
            ff = [0]
        elif ff == 1:
            ff = [-1, 0, 1]
        else:
            raise DSPFatal(
                "The penalized coefficient for the ref matrix must be 0 or 1"
            )
        for i in ff:
            ref_mat += np.outer(
                reference[flo + i : fhi + i], reference[flo + i : fhi + i]
            )
            ref_sig += reference[flo + i : fhi + i]
        ref_mat /= len(ff)

        # filter calculation
        mat = a1 * noise_mat + a2 * ref_mat + a3 * np.ones([length, length])
        x = np.linalg.solve(mat, ref_sig)
        y = np.convolve(reference, np.flip(x), mode="valid")
        maxy = np.amax(y)
        x /= maxy
        y /= maxy

    @guvectorize(
        ["void(float32[:], float32[:])", "void(float64[:], float64[:])"],
        "(n),(m)",
        **nb_kwargs(
            cache=False,
            forceobj=True,
        ),
    )
    def dplms_out(w_in: np.ndarray, w_out: np.ndarray) -> None:
        """
        Parameters
        ----------
        w_in
            the input waveform.
        w_out
            the filtered waveform.
        """

        w_out[:] = np.nan

        if np.isnan(w_in).any():
            return

        if len(x) > len(w_in):
            raise DSPFatal("The filter is longer than the input waveform")

        w_out[:] = np.convolve(w_in, np.flip(x), "valid")

    return dplms_out



"""
pygama convenience functions for 1D histograms.

1D hists in pygama require 3 things available from all implementations
of 1D histograms of numerical data in python: hist, bins, and var:
- hist: an array of histogram values
- bins: an array of bin edges
- var: an array of variances in each bin
If var is not provided, pygama assuems that the hist contains "counts" with
variance = counts (Poisson stats)

These are just convenience functions, provided for your convenience. Hopefully
they will help you if you need to do something trickier than is provided (e.g.
2D hists).
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rcParams

import pygama.math.utils as pgu


def get_hist(data, bins=None, range=None, dx=None, wts=None):
    """return hist, bins, var after binning data

    This is just a wrapper for numpy.histogram, with optional weights for each
    element and proper computing of variances.

    Note: there are no overflow / underflow bins.

    Available binning methods:

      - Default (no binning arguments) : 100 bins over an auto-detected range
      - bins=N, range=(x_lo, x_hi) : N bins over the specified range (or leave
        range=None for auto-detected range)
      - bins=[str] : use one of np.histogram's automatic binning algorithms
      - bins=bin_edges_array : array lower bin edges, supports non-uniform binning
      - dx=dx, range=(x_lo, x_hi): bins of width dx over the specified range.
        Note: dx overrides the bins argument!

    Parameters
    ----------
    data : array like
        The array of data to be histogrammed
    bins : int, array, or str (optional)
        int: the number of bins to be used in the histogram
        array: an array of bin edges to use
        str: the name of the np.histogram automatic binning algorithm to use
        If not provided, np.histogram's default auto-binning routine is used
    range : tuple (float, float) (optional)
        (x_lo, x_high) is the tuple of low and high x values to uses for the
        very ends of the bin range. If not provided, np.histogram chooses the
        ends based on the data in data
    dx : float (optional)
        Specifies the bin width. Overrides "bins" if both arguments are present
    wts : float or array like (optional)
        Array of weights for each bin. For example, if you want to divide all
        bins by a time T to get the bin contents in count rate, set wts = 1/T.
        Variances will be computed for each bin that appropriately account for
        each data point's weighting.

    Returns
    -------
    hist : array
        the values in each bin of the histogram
    bins : array
        an array of bin edges: bins[i] is the lower edge of the ith bin.
        Note: it includes the upper edge of the last bin and does not
        include underflow or overflow bins. So len(bins) = len(hist) + 1
    var : array
        array of variances in each bin of the histogram
    """
    if bins is None:
        bins = 100 # override np.histogram default of just 10

    if dx is not None:
        bins = int((range[1] - range[0]) / dx)

    # bins includes left edge of first bin and right edge of all other bins
    # allow scalar weights
    if wts is not None and np.shape(wts) == (): wts = np.full_like(data, wts)
    hist, bins = np.histogram(data, bins=bins, range=range, weights=wts)

    if wts is None:
        # no weights: var = hist, but return a copy so that mods to var don't
        # modify hist.
        # Note: If you don't want a var copy, just call np.histogram()
        return hist, bins, hist.copy()
    else:
        # get the variances by binning with double the weight
        var, bins = np.histogram(data, bins=bins, weights=wts*wts)
        return hist, bins, var


def better_int_binning(x_lo=0, x_hi=None, dx=None, n_bins=None):
    """ Get a good binning for integer data.

    Guarantees an integer bin width.

    At least two of x_hi, dx, or n_bins must be provided.

    Parameters
    ----------
    x_lo : float
        Desired low x value for the binning
    x_hi : float
        Desired high x value for the binning
    dx : float
        Desired bin width
    n_bins : float
        Desired number of bins

    Returns
    -------
    x_lo: int
        int values for best x_lo
    x_hi: int
        int values for best x_hi, returned if x_hi is not None
    dx : int
        best int bin width, returned if arg dx is not None
    n_bins : int
        best int n_bins, returned if arg n_bins is not None
    """
    # process inputs
    n_Nones = int(x_hi is None) + int(dx is None) + int(n_bins is None)
    if n_Nones > 1:
        print('better_int_binning: must provide two of x_hi, dx or n_bins')
        return
    if n_Nones == 0:
        print('better_int_binning: overconstrained. Ignoring x_hi.')
        x_hi = None

    # get valid dx or n_bins
    if dx is not None:
        if dx <= 0:
            print(f'better_int_binning: invalid dx={dx}')
            return
        dx = np.round(dx)
        if dx == 0: dx = 1
    if n_bins is not None:
        if n_bins <= 0:
            print(f'better_int_binning: invalid n_bins={n_bins}')
            return
        n_bins = np.round(n_bins)

    # can already return if no x_hi
    if x_hi is None: # must have both dx and n_bins
        return int(x_lo), int(dx), int(n_bins)

    # x_hi is valid. Get a valid dx if we don't have one
    if dx is None: # must have n_bins
        dx = np.round((x_hi-x_lo)/n_bins)
    if dx == 0: dx = 1

    # Finally, build a good binning from dx
    n_bins = np.ceil((x_hi-x_lo)/dx)
    x_lo = np.floor(x_lo)
    x_hi = x_lo + n_bins*dx
    if n_bins is None: return int(x_lo), int(x_hi), int(dx)
    else: return int(x_lo), int(x_hi), int(n_bins)


def get_bin_centers(bins):
    """
    Returns an array of bin centers from an input array of bin edges.
    Works for non-uniform binning. Note: a new array is allocated

    Parameters:
    """
    return (bins[:-1] + bins[1:]) / 2.


def get_bin_widths(bins):
    """
    Returns an array of bin widths from an input array of bin edges.
    Works for non-uniform binning.
    """
    return (bins[1:] - bins[:-1])


def find_bin(x, bins):
    """
    Returns the index of the bin containing x
    Returns -1 for underflow, and len(bins) for overflow
    For uniform bins, jumps directly to the appropriate index.
    For non-uniform bins, binary search is used.
    """
    # first handle overflow / underflow
    if len(bins) == 0: return 0 # i.e. overflow
    if x < bins[0]: return -1
    if x > bins[-1]: return len(bins)

    # we are definitely in range, and there are at least 2 bin edges, one below
    # and one above x. try assuming uniform bins
    dx = bins[1]-bins[0]
    index = int(np.floor((x-bins[0])/dx))
    if bins[index] <= x and bins[index+1] > x: return index

    # bins are non-uniform: find by binary search
    return np.searchsorted(bins, x, side='right')


def range_slice(x_min, x_max, hist, bins, var=None):
    i_min = find_bin(x_min, bins)
    i_max = find_bin(x_max, bins)
    if var is not None: var = var[i_min:i_max]
    return hist[i_min:i_max], bins[i_min:i_max+1], var


def get_fwhm(hist, bins, var=None, mx=None, dmx=0, bl=0, dbl=0, method='bins_over_f', n_slope=3):
    """Estimate the FWHM of data in a histogram

    See Also
    --------
    get_fwfm : for parameters and return values
    """
    if len(bins) == len(hist):
        print("note: this function has been updated to require bins rather",
              "than bin_centers. Don't trust this result")
    return get_fwfm(0.5, hist, bins, var, mx, dmx, bl, dbl, method, n_slope)


def get_fwfm(fraction, hist, bins, var=None, mx=None, dmx=0, bl=0, dbl=0, method='bins_over_f', n_slope=3):
    """
    Estimate the full width at some fraction of the max of data in a histogram

    Typically used by sending slices around a peak. Searches about argmax(hist)
    for the peak to fall by [fraction] from mx to bl

    Parameters
    ----------
    fraction : float
        The fractional amplitude at which to evaluate the full width
    hist : array-like
        The histogram data array containing the peak
    bins : array-like
        An array of bin edges for the histogram
    var : array-like (optional)
        An array of histogram variances. Used with the 'fit_slopes' method
    mx : float or tuple(float, float) (optional)
        The value to use for the max of the peak. If None, np.amax(hist) is
        used.
    dmx : float (optional)
        The uncertainty in mx
    bl : float or tuple (float, float) (optional)
        Used to specify an offset from which to estimate the FWFM.
    dbl : float (optional)
        The uncertainty in the bl
    method : string
        'bins_over_f' : the simplest method: just take the difference in the bin
            centers that are over [fraction] of max. Only works for high stats and
            FWFM/bin_width >> 1
        'interpolate' : interpolate between the bins that cross the [fration]
            line.  Works well for high stats and a reasonable number of bins.
            Uncertainty incorporates var, if provided.
        'fit_slopes' : fit over n_slope bins in the vicinity of the FWFM and
            interpolate to get the fractional crossing point. Works okay even
            when stats are moderate but requires enough bins that dx traversed
            by n_slope bins is approximately linear. Incorporates bin variances
            in fit and final uncertainties if provided.
    n_slope : int
        DOCME

    Returns
    -------
    fwfm, dfwfm : float, float
        fwfm: the full width at [fraction] of the maximum above bl
        dfwfm: the uncertainty in fwfm

    Examples
    --------
    >>> import pygama.analysis.histograms as pgh
    >>> from numpy.random import normal
    >>> hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5,5))
    >>> pgh.get_fwfm(0.5, hist, bins, var, method='bins_over_f')
    (2.2, 0.15919638684132664) # may vary

    >>> pgh.get_fwfm(0.5, hist, bins, var, method='interpolate')
    (2.2041666666666666, 0.09790931254396479) # may vary

    >>> pgh.get_fwfm(0.5, hist, bins, var, method='fit_slopes')
    (2.3083363869003466, 0.10939486522749278) # may vary
    """

    # find bins over [fraction]
    if mx is None:
        mx = np.amax(hist)
        if var is not None and dmx == 0:
            dmx = np.sqrt(var[np.argmax(hist)])
    idxs_over_f = hist > (bl + fraction * (mx-bl))

    # argmax will return the index of the first occurrence of a maximum
    # so we can use it to find the first and last time idxs_over_f is "True"
    bin_lo = np.argmax(idxs_over_f)
    bin_hi = len(idxs_over_f) - np.argmax(idxs_over_f[::-1])
    bin_centers = get_bin_centers(bins)

    # precalc dheight: uncertainty in height used as the threshold
    dheight2 = (fraction*dmx)**2 + ((1-fraction)*dbl)**2

    if method == 'bins_over_f':
        # the simplest method: just take the difference in the bin centers
        fwfm = bin_centers[bin_hi] - bin_centers[bin_lo]

        # compute rough uncertainty as [bin width] (+) [dheight / slope]
        dx = bin_centers[bin_lo] - bin_centers[bin_lo-1]
        dy = hist[bin_lo] - hist[bin_lo-1]
        if dy == 0: dy = (hist[bin_lo+1] - hist[bin_lo-2])/3
        dfwfm2 = dx**2 + dheight2 * (dx/dy)**2
        dx = bin_centers[bin_hi+1] - bin_centers[bin_hi]
        dy = hist[bin_hi] - hist[bin_hi+1]
        if dy == 0: dy = (hist[bin_hi-1] - hist[bin_hi+2])/3
        dfwfm2 += dx**2 + dheight2 * (dx/dy)**2
        return fwfm, np.sqrt(dfwfm2)

    elif method == 'interpolate':
        # interpolate between the two bins that cross the [fraction] line
        # works well for high stats
        if bin_lo < 1 or bin_hi >= len(hist)-1:
            print(f"get_fwhm: can't interpolate ({bin_lo}, {bin_hi})")
            return 0, 0

        val_f = bl + fraction*(mx-bl)

        # x_lo
        dx = bin_centers[bin_lo] - bin_centers[bin_lo-1]
        dhf = val_f - hist[bin_lo-1]
        dh = hist[bin_lo] - hist[bin_lo-1]
        x_lo = bin_centers[bin_lo-1] + dx * dhf/dh
        # uncertainty
        dx2_lo = 0
        if var is not None:
            dx2_lo = (dhf/dh)**2 * var[bin_lo] + ((dh-dhf)/dh)**2 * var[bin_lo-1]
            dx2_lo *= (dx/dh)**2
        dDdh = -dx/dh

        # x_hi
        dx = bin_centers[bin_hi+1] - bin_centers[bin_hi]
        dhf = hist[bin_hi] - val_f
        dh = hist[bin_hi] - hist[bin_hi+1]
        if dh == 0:
            raise ValueError(f"get_fwhm: interpolation failed, dh == 0")
        x_hi = bin_centers[bin_hi] + dx * dhf/dh
        if x_hi < x_lo:
            raise ValueError(f"get_fwfm: interpolation produced negative fwfm")
        # uncertainty
        dx2_hi = 0
        if var is not None:
            dx2_hi = (dhf/dh)**2 * var[bin_hi+1] + ((dh-dhf)/dh)**2 * var[bin_hi]
            dx2_hi *= (dx/dh)**2
        dDdh += dx/dh

        return x_hi - x_lo, np.sqrt(dx2_lo + dx2_hi + dDdh**2 * dheight2)

    elif method == 'fit_slopes':
        # evaluate the [fraction] point on a line fit to n_slope bins near the crossing.
        # works okay even when stats are moderate
        val_f = bl + fraction*(mx-bl)

        # x_lo
        i_0 = bin_lo - int(np.floor(n_slope/2))
        if i_0 < 0:
            print(f"get_fwfm: fit slopes failed")
            return 0, 0
        i_n = i_0 + n_slope
        wts = None if var is None else 1/np.sqrt(var[i_0:i_n]) #fails for any var = 0
        wts = [w if w != np.inf else 0 for w in wts]

        try:
            (m, b), cov = np.polyfit(bin_centers[i_0:i_n], hist[i_0:i_n], 1, w=wts, cov='unscaled')
        except np.linalg.LinAlgError:
            print(f"get_fwfm: LinAlgError")
            return 0, 0
        x_lo = (val_f-b)/m
        #uncertainty
        dxl2 = cov[0,0]/m**2 + (cov[1,1] + dheight2)/(val_f-b)**2 + 2*cov[0,1]/(val_f-b)/m
        dxl2 *= x_lo**2

        # x_hi
        i_0 = bin_hi - int(np.floor(n_slope/2)) + 1
        if i_0 == len(hist):
            print(f"get_fwfm: fit slopes failed")
            return 0, 0

        i_n = i_0 + n_slope
        wts = None if var is None else 1/np.sqrt(var[i_0:i_n])
        wts = [w if w != np.inf else 0 for w in wts]
        try:
            (m, b), cov = np.polyfit(bin_centers[i_0:i_n], hist[i_0:i_n], 1, w=wts, cov='unscaled')
        except np.linalg.LinAlgError:
            print(f"get_fwfm: LinAlgError")
            return 0, 0
        x_hi = (val_f-b)/m
        if x_hi < x_lo:
            print(f"get_fwfm: fit slopes produced negative fwfm")
            return 0, 0

        #uncertainty
        dxh2 = cov[0,0]/m**2 + (cov[1,1] + dheight2)/(val_f-b)**2 + 2*cov[0,1]/(val_f-b)/m
        dxh2 *= x_hi**2

        return x_hi - x_lo, np.sqrt(dxl2 + dxh2)

    else:
        print(f"get_fwhm: unrecognized method {method}")
        return 0, 0


def plot_hist(hist, bins, var=None, show_stats=False, stats_hloc=0.75, stats_vloc=0.85, fill=False, fillcolor='r', fillalpha=0.2, **kwargs):
    """
    plot a step histogram, with optional error bars
    """
    if fill:
        # the concat calls get the steps to draw correctly at the range boundaries
        # where="post" tells plt to draw the step y[i] between x[i] and x[i+1]
        save_color = None
        if 'color' in kwargs: save_color = kwargs.pop('color')
        elif 'c' in kwargs: save_color = kwargs.pop('c')
        plt.fill_between(np.concatenate(([bins[0]], bins)), np.concatenate(([0], hist, [0])), step="post", color=fillcolor, alpha=fillalpha, **kwargs)
        if save_color is not None: kwargs['color'] = save_color
    if var is None:
        # the concat calls get the steps to draw correctly at the range boundaries
        # where="post" tells plt to draw the step y[i] between x[i] and x[i+1]
        plt.step(np.concatenate(([bins[0]], bins)), np.concatenate(([0], hist, [0])), where="post", **kwargs)
    else:
        plt.errorbar(get_bin_centers(bins), hist,
                     xerr=get_bin_widths(bins) / 2, yerr=np.sqrt(var),
                     fmt='none', **kwargs)
    if show_stats is True:
        bin_centers = get_bin_centers(bins)
        N = np.sum(hist)
        if N <= 1:
            print("can't compute sigma for N =", N)
            return
        mean = np.sum(hist*bin_centers)/N
        x2ave = np.sum(hist*bin_centers*bin_centers)/N
        stddev = np.sqrt(N/(N-1) * (x2ave - mean*mean))
        dmean = stddev/np.sqrt(N)

        mean, dmean = pgu.get_formatted_stats(mean, dmean, 2)
        stats = f'$\\mu={mean} \\pm {dmean}$\n$\\sigma={stddev:#.3g}$'
        stats_fontsize = rcParams['legend.fontsize']
        plt.text(stats_hloc, stats_vloc, stats, transform=plt.gca().transAxes, fontsize = stats_fontsize)


def get_gaussian_guess(hist, bins):
    """
    given a hist, gives guesses for mu, sigma, and amplitude
    """
    if len(bins) == len(hist):
        print("note: this function has been updated to require bins rather",
              "than bin_centers. Don't trust this result")

    max_idx = np.argmax(hist)
    guess_e = (bins[max_idx] + bins[max_idx])/2 # bin center
    guess_amp = hist[max_idx]

    # find 50% amp bounds on both sides for a FWHM guess
    guess_sigma = get_fwhm(hist, bins)[0] / 2.355  # FWHM to sigma
    guess_area = guess_amp * guess_sigma * np.sqrt(2 * np.pi)

    return (guess_e, guess_sigma, guess_area)



"""
Base classes for streaming data.
"""
from __future__ import annotations

import fnmatch
import logging
from abc import ABC, abstractmethod

from .raw_buffer import RawBuffer, RawBufferLibrary, RawBufferList

log = logging.getLogger(__name__)


class DataStreamer(ABC):
    """Base clase for data streams.

    Provides a uniform interface for streaming, e.g.:

    >>> header = ds.open_stream(stream_name)
    >>> for chunk in ds: do_something(chunk)

    Also provides default management of the :class:`.RawBufferLibrary` used for
    data reading: allocation (if needed), configuration (to match the stream)
    and fill level checking.  Derived classes must define the functions
    :meth:`.get_decoder_list`, :meth:`.open_stream`, and :meth:`.read_packet`;
    see below.
    """

    def __init__(self) -> None:
        self.rb_lib = None
        self.chunk_mode = None
        self.n_bytes_read = 0
        self.any_full = False
        self.packet_id = 0

    @abstractmethod
    def open_stream(
        self,
        stream_name: str,
        rb_lib: RawBufferLibrary = None,
        buffer_size: int = 8192,
        chunk_mode: str = "any_full",
        out_stream: str = "",
    ) -> tuple[list[RawBuffer], int]:
        r"""Open and initialize a data stream.

        Open the stream, read in the header, set up the buffers.

        Call ``super().initialize([args])`` from derived class after loading
        header info to run this default version that sets up buffers in
        `rb_lib` using the stream's decoders.

        Notes
        -----
        this default version has no actual return value! You must overload this
        function, set :attr:`self.n_bytes_read` to the header packet size, and
        return the header data.

        Parameters
        ----------
        stream_name
            typically a filename or e.g. a port for streaming.
        rb_lib
            a library of buffers for readout from the data stream. `rb_lib`
            will have its LGDO's initialized during this function.
        buffer_size
            length of buffers to be read out in :meth:`read_chunk` (for buffers
            with variable length).
        chunk_mode : 'any_full', 'only_full' or 'single_packet'
            sets the mode use for :meth:`read_chunk`.
        out_stream
            optional name of output stream for default `rb_lib` generation.

        Returns
        -------
        header_data
            header_data is a list of :class:`.RawBuffer`\ 's containing any
            file header data, ready for writing to file or further processing.
            It's not a :class:`.RawBufferList` since the buffers may have a
            different format.
        """
        # call super().initialize([args]) to run this default code
        # after loading header info, then follow it with the return call.

        # store chunk mode
        self.chunk_mode = chunk_mode

        # prepare rb_lib -- its lgdo's should still be uninitialized
        if rb_lib is None:
            rb_lib = self.build_default_rb_lib(out_stream=out_stream)
        self.rb_lib = rb_lib

        # now initialize lgdo's for raw buffers
        decoders = self.get_decoder_list()
        dec_names = []
        for decoder in decoders:
            dec_name = type(decoder).__name__

            # set up wildcard decoder buffers
            if dec_name not in rb_lib:
                if "*" not in rb_lib:
                    continue  # user didn't want this decoder
                rb_lib[dec_name] = RawBufferList()
                dec_key = dec_name
                if dec_key.endswith("Decoder"):
                    dec_key = dec_key.removesuffix("Decoder")
                out_name = rb_lib["*"][0].out_name.format(name=dec_key)
                out_stream = rb_lib["*"][0].out_stream.format(name=dec_key)
                proc_spec = rb_lib["*"][0].proc_spec
                key_lists = decoder.get_key_lists()
                for ii, key_list in enumerate(key_lists):
                    this_name = out_name
                    if len(key_lists) > 1:
                        if len(key_list) == 1:
                            this_name = f"{out_name}_{key_list[0]}"
                        else:
                            this_name = f"{out_name}_{ii}"
                    rb = RawBuffer(
                        key_list=key_list,
                        out_stream=out_stream,
                        out_name=this_name,
                        proc_spec=proc_spec,
                    )
                    rb_lib[dec_name].append(rb)

            # dec_name is in rb_lib: store the name, and initialize its buffer lgdos
            dec_names.append(dec_name)

            # set up wildcard key buffers
            for rb in rb_lib[dec_name]:
                if (
                    len(rb.key_list) == 1
                    and isinstance(rb.key_list[0], str)
                    and "*" in rb.key_list[0]
                ):
                    matched_key_lists = []
                    for key_list in decoder.get_key_lists():
                        # special case: decoders without keys
                        if rb.key_list[0] == "*" and key_list == [None]:
                            matched_key_lists.append(key_list)
                            continue
                        key_type = type(key_list[0])
                        for ik in range(len(key_list)):
                            key_list[ik] = str(key_list[ik])
                        matched_keys = fnmatch.filter(key_list, rb.key_list[0])
                        if len(matched_keys) > 1:
                            for ik in range(len(matched_keys)):
                                matched_keys[ik] = key_type(key_list[ik])
                            matched_key_lists.append(matched_keys)
                    if len(matched_key_lists) == 0:
                        log.warning(
                            f"no matched keys for key_list {rb.key_list[0]} in {dec_name}.{rb.out_name}"
                        )
                        continue
                    rb.key_list = sum(matched_key_lists, [])
            keyed_name_rbs = []
            ii = 0
            while ii < len(rb_lib[dec_name]):
                if "{key" in rb_lib[dec_name][ii].out_name:
                    keyed_name_rbs.append(rb_lib[dec_name].pop(ii))
                else:
                    ii += 1
            for rb in keyed_name_rbs:
                for key in rb.key_list:
                    # keys can be strs or ints; try as-is, but can get a
                    # ValueError e.g. when using a wildcard with int keys. In
                    # that case, switch to the other type and try again
                    try:
                        expanded_name = rb.out_name.format(key=key)
                    except ValueError:
                        if isinstance(key, str):
                            key = int(key)
                        else:
                            key = str(key)
                        expanded_name = rb.out_name.format(key=key)

                    new_rb = RawBuffer(
                        key_list=[key],
                        out_stream=rb.out_stream,
                        out_name=expanded_name,
                        proc_spec=rb.proc_spec,
                    )
                    rb_lib[dec_name].append(new_rb)

            for rb in rb_lib[dec_name]:
                # use the first available key
                key = rb.key_list[0] if len(rb.key_list) > 0 else None
                rb.lgdo = decoder.make_lgdo(key=key, size=buffer_size)

        # make sure there were no entries in rb_lib that weren't among the
        # decoders. If so, just emit a warning and continue.
        if "*" in rb_lib:
            rb_lib.pop("*")
        for dec_name in rb_lib.keys():
            if dec_name not in dec_names:
                log.warning(f"no decoder named '{dec_name}' requested by rb_lib")

    @abstractmethod
    def close_stream(self) -> None:
        """Close this data stream.

        .. note::
            Needs to be overloaded.
        """
        pass

    @abstractmethod
    def read_packet(self) -> bool:
        """Reads a single packet's worth of data in to the :class:`.RawBufferLibrary`.

        Needs to be overloaded. Gets called by :meth:`.read_chunk` Needs to
        update :attr:`self.any_full` if any buffers would possibly over-fill on
        the next read. Needs to update :attr:`self.n_bytes_read` too.

        Returns
        -------
        still_has_data
            returns `True` while there is still data to read.
        """
        return True

    def read_chunk(
        self,
        chunk_mode_override: str = None,
        rp_max: int = 1000000,
        clear_full_buffers: bool = True,
    ) -> tuple[list[RawBuffer], int]:
        """Reads a chunk of data into raw buffers.

        Reads packets until at least one buffer is too full to perform another
        read. Default version just calls :meth:`.read_packet` over and over.
        Overload as necessary.

        Notes
        -----
        user is responsible for resetting / clearing the raw buffers prior to
        calling :meth:`.read_chunk` again.

        Parameters
        ----------
        chunk_mode_override : 'any_full', 'only_full' or 'single_packet'
            - ``None`` : do not override self.chunk_mode
            - ``any_full`` : returns all raw buffers with data as soon as any one
              buffer gets full
            - ``only_full`` : returns only those raw buffers that became full (or
              nearly full) during the read. This minimizes the number of write calls.
            - ``single_packet`` : returns all raw buffers with data after a single
              read is performed. This is useful for streaming data out as soon
              as it is read in (e.g. for diagnostics or in-line analysis).
        rp_max
            maximum number of packets to read before returning anyway, even if
            one of the other conditions is not met.
        clear_full_buffers
            automatically clear any buffers that report themselves as being
            full prior to reading the chunk. Set to `False` if clearing
            manually for a minor speed-up.

        Returns
        -------
        chunk_list : list of RawBuffers, int
            chunk_list is the list of RawBuffers with data ready for writing to
            file or further processing. The list contains all buffers with data
            or just all full buffers depending on the flag full_only.  Note
            chunk_list is not a RawBufferList since the RawBuffers inside may
            not all have the same structure
        """

        if clear_full_buffers:
            self.rb_lib.clear_full()
        self.any_full = False

        chunk_mode = (
            self.chunk_mode if chunk_mode_override is None else chunk_mode_override
        )

        read_one_packet = chunk_mode == "single_packet"
        only_full = chunk_mode == "only_full"

        n_packets = 0
        still_has_data = True
        while True:
            still_has_data = self.read_packet()
            if not still_has_data:
                break
            n_packets += 1
            if read_one_packet or n_packets > rp_max:
                break
            if self.any_full:
                break

        # send back all rb's with data if we finished reading
        if not still_has_data:
            only_full = False

        list_of_rbs = []
        for rb_list in self.rb_lib.values():
            for rb in rb_list:
                if not only_full:  # any_full or read_one_packet
                    if rb.loc > 0:
                        list_of_rbs.append(rb)
                elif rb.is_full():
                    list_of_rbs.append(rb)
        if not still_has_data:
            log.debug(f"decoding complete. flushing {len(list_of_rbs)} buffers")
        return list_of_rbs

    @abstractmethod
    def get_decoder_list(self) -> list:
        """Returns a list of decoder objects for this data stream.

        Notes
        -----
        Needs to be overloaded. Gets called during :meth:`.open_stream`.
        """
        return []

    def build_default_rb_lib(self, out_stream: str = "") -> RawBufferLibrary:
        """Build the most basic :class:`~.RawBufferLibrary` that will work for
        this stream.

        A :class:`.RawBufferList` containing a single :class:`.RawBuffer` is
        built for each decoder name returned by :meth:`.get_decoder_list`. Each
        buffer's `out_name` is set to the decoder name. The LGDO's do not get
        initialized.
        """
        rb_lib = RawBufferLibrary()
        decoders = self.get_decoder_list()
        if len(decoders) == 0:
            log.warning(
                f"no decoders returned by get_decoder_list() for {type(self).__name__}"
            )
            return rb_lib
        for decoder in decoders:
            dec_name = type(decoder).__name__
            dec_key = dec_name
            if dec_key.endswith("Decoder"):
                dec_key = dec_key.removesuffix("Decoder")
            key_lists = decoder.get_key_lists()
            for ii, key_list in enumerate(key_lists):
                this_name = dec_key
                if len(key_lists) > 1:
                    if len(key_list) == 1:
                        this_name = f"{dec_key}_{key_list[0]}"
                    else:
                        this_name = f"{dec_key}_{ii}"
                rb = RawBuffer(
                    key_list=key_list, out_stream=out_stream, out_name=this_name
                )
                if dec_name not in rb_lib:
                    rb_lib[dec_name] = RawBufferList()
                rb_lib[dec_name].append(rb)
        return rb_lib



"""Implements utilities for LEGEND Data Objects."""
from __future__ import annotations

import glob
import logging
import os
import string

import numpy as np

from .. import lgdo

log = logging.getLogger(__name__)


def get_element_type(obj: object) -> str:
    """Get the LGDO element type of a scalar or array.

    For use in LGDO datatype attributes.

    Parameters
    ----------
    obj
        if a ``str``, will automatically return ``string`` if the object has
        a :class:`numpy.dtype`, that will be used for determining the element
        type otherwise will attempt to case the type of the object to a
        :class:`numpy.dtype`.

    Returns
    -------
    element_type
        A string stating the determined element type of the object.
    """

    # special handling for strings
    if isinstance(obj, str):
        return "string"

    # the rest use dtypes
    dt = obj.dtype if hasattr(obj, "dtype") else np.dtype(type(obj))
    kind = dt.kind

    if kind == "b":
        return "bool"
    if kind == "V":
        return "blob"
    if kind in ["i", "u", "f"]:
        return "real"
    if kind == "c":
        return "complex"
    if kind in ["S", "U"]:
        return "string"

    # couldn't figure it out
    raise ValueError(
        "cannot determine lgdo element_type for object of type", type(obj).__name__
    )


def copy(obj: lgdo.LGDO, dtype: np.dtype = None) -> lgdo.LGDO:
    """Return a copy of an LGDO.

    Parameters
    ----------
    obj
        the LGDO to be copied.
    dtype
        NumPy dtype to be used for the copied object.

    """
    if dtype is None:
        dtype = obj.dtype

    if isinstance(obj, lgdo.Array):
        return lgdo.Array(
            np.array(obj.nda, dtype=dtype, copy=True), attrs=dict(obj.attrs)
        )

    if isinstance(obj, lgdo.VectorOfVectors):
        return lgdo.VectorOfVectors(
            flattened_data=copy(obj.flattened_data, dtype=dtype),
            cumulative_length=copy(obj.cumulative_length),
            attrs=dict(obj.attrs),
        )

    else:
        raise ValueError(f"copy of {type(obj)} not supported")


def parse_datatype(datatype: str) -> tuple[str, tuple[int, ...], str | list[str]]:
    """Parse datatype string and return type, dimensions and elements.

    Parameters
    ----------
    datatype
        a LGDO-formatted datatype string.

    Returns
    -------
    element_type
        the datatype name dims if not ``None``, a tuple of dimensions for the
        LGDO. Note this is not the same as the NumPy shape of the underlying
        data object. See the LGDO specification for more information. Also see
        :class:`.ArrayOfEqualSizedArrays` and
        :meth:`.lh5_store.LH5Store.read_object` for example code elements for
        numeric objects, the element type for struct-like  objects, the list of
        fields in the struct.
    """
    if "{" not in datatype:
        return "scalar", None, datatype

    # for other datatypes, need to parse the datatype string
    from parse import parse

    datatype, element_description = parse("{}{{{}}}", datatype)
    if datatype.endswith(">"):
        datatype, dims = parse("{}<{}>", datatype)
        dims = [int(i) for i in dims.split(",")]
        return datatype, tuple(dims), element_description
    else:
        return datatype, None, element_description.split(",")


def expand_vars(expr: str, substitute: dict[str, str] = None) -> str:
    """Expand (environment) variables.

    Note
    ----
    Malformed variable names and references to non-existing variables are left
    unchanged.

    Parameters
    ----------
    expr
        string expression, which may include (environment) variables prefixed by
        ``$``.
    substitute
        use this dictionary to substitute variables. Environment variables take
        precedence.
    """
    if substitute is None:
        substitute = {}

    # expand env variables first
    # then try using provided mapping
    return string.Template(os.path.expandvars(expr)).safe_substitute(substitute)


def expand_path(
    path: str,
    substitute: dict[str, str] = None,
    list: bool = False,
    base_path: str = None,
) -> str | list:
    """Expand (environment) variables and wildcards to return absolute paths.

    Parameters
    ----------
    path
        name of path, which may include environment variables and wildcards.
    list
        if ``True``, return a list. If ``False``, return a string; if ``False``
        and a unique file is not found, raise an exception.
    substitute
        use this dictionary to substitute variables. Environment variables take
        precedence.
    base_path
        name of base path. Returned paths will be relative to base.

    Returns
    -------
    path or list of paths
        Unique absolute path, or list of all absolute paths
    """
    if base_path is not None and base_path != "":
        base_path = os.path.expanduser(os.path.expandvars(base_path))
        path = os.path.join(base_path, path)

    # first expand variables
    _path = expand_vars(path, substitute)

    # then expand wildcards
    paths = glob.glob(os.path.expanduser(_path))

    if base_path is not None and base_path != "":
        paths = [os.path.relpath(p, base_path) for p in paths]

    if not list:
        if len(paths) == 0:
            raise FileNotFoundError(f"could not find path matching {path}")
        elif len(paths) > 1:
            raise FileNotFoundError(f"found multiple paths matching {path}")
        else:
            return paths[0]
    else:
        return paths



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->(n)",
    **nb_kwargs,
)
def bl_subtract(w_in: np.ndarray, a_baseline: float, w_out: np.ndarray) -> None:
    """Subtract the constant baseline from the entire waveform.

    Parameters
    ----------
    w_in
        the input waveform.
    a_baseline
        the baseline value to subtract.
    w_out
        the output waveform with the baseline subtracted.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_bl": {
            "function": "bl_subtract",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "baseline", "wf_bl"],
            "unit": "ADC"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any() or np.isnan(a_baseline):
        return

    w_out[:] = w_in[:] - a_baseline



from pygama.raw.compass.compass_streamer import CompassStreamer
from pygama.raw.raw_buffer import RawBuffer, RawBufferList


def test_default_rb_lib(lgnd_test_data):
    streamer = CompassStreamer(
        lgnd_test_data.get_path("compass/compass_test_data_settings.xml"),
    )
    streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"), buffer_size=6
    )
    rb_lib = streamer.build_default_rb_lib()
    assert "CompassHeaderDecoder" in rb_lib.keys()
    assert "CompassEventDecoder" in rb_lib.keys()
    assert rb_lib["CompassHeaderDecoder"][0].out_name == "CompassHeader"
    assert rb_lib["CompassEventDecoder"][0].out_name == "CompassEvent"
    assert rb_lib["CompassEventDecoder"][0].key_list == [0, 1]
    streamer.close_stream()


def test_open_stream(lgnd_test_data):
    streamer = CompassStreamer()
    res = streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"), buffer_size=6
    )
    assert isinstance(res[0], RawBuffer)
    assert streamer.header is not None  # header object is instantiated
    assert streamer.packet_id == 0  # packet id is initialized
    assert streamer.n_bytes_read == 2  # CoMPASS header is read
    assert streamer.event_rbkd is not None  # dict containing event info is initialized
    streamer.close_stream()


def test_read_packet(lgnd_test_data):
    streamer = CompassStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"), buffer_size=6
    )
    init_rbytes = streamer.n_bytes_read
    assert streamer.read_packet() is True  # read was successful
    assert streamer.packet_id == 1  # packet id is incremented
    assert (
        streamer.n_bytes_read
        == init_rbytes + 25 + 2 * streamer.header["boards"]["0"]["wf_len"].value
    )
    streamer.close_stream()


def test_read_packet_partial(lgnd_test_data):
    streamer = CompassStreamer()
    rb_lib = {"CompassEventDecoder": RawBufferList()}
    rb_lib["CompassEventDecoder"].append(
        RawBuffer(key_list=range(0, 1), out_name="events")
    )

    streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"),
        rb_lib=rb_lib,
        buffer_size=6,
    )

    assert list(streamer.rb_lib.keys()) == ["CompassEventDecoder"]
    assert streamer.rb_lib["CompassEventDecoder"][0].key_list == range(0, 1)
    assert streamer.rb_lib["CompassEventDecoder"][0].out_name == "events"

    init_rbytes = streamer.n_bytes_read
    assert streamer.read_packet() is True  # read was successful
    assert streamer.packet_id == 1  # packet id is incremented
    assert (
        streamer.n_bytes_read
        == init_rbytes + 25 + 2 * streamer.header["boards"]["0"]["wf_len"].value
    )
    streamer.close_stream()


def test_read_chunk(lgnd_test_data):
    streamer = CompassStreamer()
    streamer.open_stream(
        lgnd_test_data.get_path("compass/compass_test_data.BIN"), buffer_size=6
    )
    streamer.read_chunk()
    streamer.close_stream()



"""
This module provides high-level routines for running signal processing chains
on waveform data.
"""
from __future__ import annotations

import json
import logging
import os
import sys
import time

import h5py
import numpy as np
from tqdm import tqdm

import pygama
import pygama.lgdo as lgdo
import pygama.lgdo.lh5_store as lh5
from pygama.dsp.errors import DSPFatal
from pygama.dsp.processing_chain import build_processing_chain
from pygama.lgdo.lgdo_utils import expand_path

log = logging.getLogger(__name__)


def build_dsp(
    f_raw: str,
    f_dsp: str,
    dsp_config: str | dict = None,
    lh5_tables: list[str] | str = None,
    database: str | dict = None,
    outputs: list[str] = None,
    n_max: int = np.inf,
    write_mode: str = None,
    buffer_len: int = 3200,
    block_width: int = 16,
    chan_config: dict[str, str] = None,
) -> None:
    """Convert raw-tier LH5 data into dsp-tier LH5 data by running a sequence
    of processors via the :class:`~.processing_chain.ProcessingChain`.

    Parameters
    ----------
    f_raw
        name of raw-tier LH5 file to read from.
    f_dsp
        name of dsp-tier LH5 file to write to.
    dsp_config
        :class:`dict` or name of JSON file containing
        :class:`~.processing_chain.ProcessingChain` config. See
        :func:`~.processing_chain.build_processing_chain` for details.
    lh5_tables
        list of HDF5 groups to consider in the input file. If ``None``, process
        all valid groups.
    database
        dictionary or name of JSON file containing a parameter database. See
        :func:`~.processing_chain.build_processing_chain` for details.
    outputs
        list of parameter names to write to the output file. If not provided,
        use list provided under ``"outputs"`` in the DSP configuration file.
    n_max
        number of waveforms to process.
    write_mode
        - ``None`` -- create new output file if it does not exist
        - `'r'` -- delete existing output file with same name before writing
        - `'a'` -- append to end of existing output file
        - `'u'` -- update values in existing output file
    buffer_len
        number of waveforms to read/write from/to disk at a time.
    block_width
        number of waveforms to process at a time.
    chan_config
        contains JSON DSP configuration file names for every table in
        `lh5_tables`.
    """

    if chan_config is not None:
        # clear existing output files
        if write_mode == "r":
            if os.path.isfile(f_dsp):
                os.remove(f_dsp)
            write_mode = "a"

        for tb, dsp_config in chan_config.items():
            log.debug(f"processing table: {tb} with DSP config file {dsp_config}")
            try:
                build_dsp(
                    f_raw,
                    f_dsp,
                    dsp_config,
                    [tb],
                    database,
                    outputs,
                    n_max,
                    write_mode,
                    buffer_len,
                    block_width,
                )
            except RuntimeError:
                log.debug(f"table {tb} not found")
        return

    raw_store = lh5.LH5Store()
    lh5_file = raw_store.gimme_file(f_raw, "r")
    if lh5_file is None:
        raise ValueError(f"input file not found: {f_raw}")
        return

    # if no group is specified, assume we want to decode every table in the file
    if lh5_tables is None:
        lh5_tables = lh5.ls(f_raw)
    elif isinstance(lh5_tables, str):
        lh5_tables = [lh5_tables]
    elif not (
        hasattr(lh5_tables, "__iter__")
        and all(isinstance(el, str) for el in lh5_tables)
    ):
        raise RuntimeError("lh5_tables must be None, a string, or a list of strings")

    # check if group points to raw data; sometimes 'raw' is nested, e.g g024/raw
    for i, tb in enumerate(lh5_tables):
        if (
            "raw" not in tb
            and not isinstance(raw_store.gimme_file(lh5_file, "r")[tb], h5py.Dataset)
            and lh5.ls(lh5_file, f"{tb}/raw")
        ):
            lh5_tables[i] = f"{tb}/raw"
        elif not lh5.ls(lh5_file, tb):
            del lh5_tables[i]

    if len(lh5_tables) == 0:
        raise RuntimeError(f"could not find any valid LH5 table in {f_raw}")

    # get the database parameters. For now, this will just be a dict in a json
    # file, but eventually we will want to interface with the metadata repo
    if isinstance(database, str):
        with open(expand_path(database)) as db_file:
            database = json.load(db_file)

    if database and not isinstance(database, dict):
        raise ValueError("input database is not a valid JSON file or dict")

    if write_mode is None and os.path.isfile(f_dsp):
        raise FileExistsError(
            f"output file {f_dsp} exists. Set the 'write_mode' keyword"
        )

    # clear existing output files
    if write_mode == "r":
        if os.path.isfile(f_dsp):
            os.remove(f_dsp)

    # write processing metadata
    dsp_info = lgdo.Struct()
    dsp_info.add_field("timestamp", lgdo.Scalar(np.uint64(time.time())))
    dsp_info.add_field("python_version", lgdo.Scalar(sys.version))
    dsp_info.add_field("numpy_version", lgdo.Scalar(np.version.version))
    dsp_info.add_field("h5py_version", lgdo.Scalar(h5py.version.version))
    dsp_info.add_field("hdf5_version", lgdo.Scalar(h5py.version.hdf5_version))
    dsp_info.add_field("pygama_version", lgdo.Scalar(pygama.__version__))

    # loop over tables to run DSP on
    for tb in lh5_tables:
        # load primary table and build processing chain and output table
        tot_n_rows = raw_store.read_n_rows(tb, f_raw)
        if n_max and n_max < tot_n_rows:
            tot_n_rows = n_max

        chan_name = tb.split("/")[0]
        db_dict = database.get(chan_name) if database else None
        tb_name = tb.replace("/raw", "/dsp")

        write_offset = 0
        raw_store.gimme_file(f_dsp, "a")
        if write_mode == "a" and lh5.ls(f_dsp, tb_name):
            write_offset = raw_store.read_n_rows(tb_name, f_dsp)

        # Main processing loop
        lh5_it = lh5.LH5Iterator(f_raw, tb, buffer_len=buffer_len)
        proc_chain = None
        for lh5_in, start_row, n_rows in lh5_it:
            # Initialize
            if proc_chain is None:
                proc_chain, lh5_it.field_mask, tb_out = build_processing_chain(
                    lh5_in, dsp_config, db_dict, outputs, block_width
                )
                if log.getEffectiveLevel() >= logging.INFO:
                    progress_bar = tqdm(
                        desc=f"Processing table {tb}",
                        total=tot_n_rows,
                        delay=2,
                        unit=" rows",
                    )

            n_rows = min(tot_n_rows - start_row, n_rows)
            try:
                proc_chain.execute(0, n_rows)
            except DSPFatal as e:
                # Update the wf_range to reflect the file position
                e.wf_range = f"{e.wf_range[0]+start_row}-{e.wf_range[1]+start_row}"
                raise e

            raw_store.write_object(
                obj=tb_out,
                name=tb_name,
                lh5_file=f_dsp,
                n_rows=n_rows,
                wo_mode="o" if write_mode == "u" else "a",
                write_start=write_offset + start_row,
            )

            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar.update(n_rows)

            if start_row + n_rows >= tot_n_rows:
                break

        if log.getEffectiveLevel() >= logging.INFO:
            progress_bar.close()

    raw_store.write_object(dsp_info, "dsp_info", f_dsp, wo_mode="o")



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, char, float32[:])",
        "void(float64[:], float64, char, float64[:])",
    ],
    "(n),(),()->()",
    **nb_kwargs,
)
def fixed_time_pickoff(w_in: np.ndarray, t_in: float, mode_in: np.int8, a_out: float):
    """Pick off the waveform value at the provided time.

    For non-integral times, interpolate between samples using the method
    selected using `mode_in`. If the provided index `t_in` is out of range,
    return :any:`numpy.nan`.

    Parameters
    ----------
    w_in
        the input waveform.
    t_in
        the waveform index to pick off.
    mode_in
        character selecting which interpolation method to use. Note this
        must be passed as a ``int8``, e.g. ``ord('i')``. Options:

        * ``i`` -- integer `t_in`; equivalent to
          :func:`~.dsp.processors.fixed_sample_pickoff`
        * ``n`` -- nearest-neighbor interpolation; defined at all values,
          but not continuous
        * ``f`` -- floor, or value at previous neighbor; defined at all
          values but not continuous
        * ``c`` -- ceiling, or value at next neighbor; defined at all values,
          but not continuous
        * ``l`` -- linear interpolation; continuous at all values, but not
          differentiable
        * ``h`` -- Hermite cubic spline interpolation; continuous and
          differentiable at all values but not twice-differentiable
        * ``s`` -- natural cubic spline interpolation; continuous and
          twice-differentiable at all values. This method is much slower
          than the others because it utilizes the entire input waveform!
    a_out
        the output pick-off value.

    Examples
    --------
    .. code-block :: json

        "trapEftp": {
            "function": "fixed_time_pickoff",
            "module": "pygama.dsp.processors",
            "args": ["wf_trap", "tp_0+10*us", "'h'", "trapEftp"],
            "unit": "ADC",
        }
    """
    a_out[0] = np.nan

    if np.isnan(w_in).any() or np.isnan(t_in):
        return

    if t_in < 0 or t_in > len(w_in) - 1:
        return

    i_in = int(t_in)
    if i_in == t_in:
        a_out[0] = w_in[i_in]
        return

    t0 = t_in - i_in
    t1 = 1 - t0

    if chr(mode_in) == "i":  # Index
        raise DSPFatal("fixed_time_pickoff requires integer t_in when using mode 'i'")
    elif chr(mode_in) == "n":  # Nearest-neighbor
        a_out[0] = w_in[i_in] if t0 < 0.5 else w_in[i_in + 1]
    elif chr(mode_in) == "f":  # Floor
        a_out[0] = w_in[i_in]
    elif chr(mode_in) == "c":  # Ceiling
        a_out[0] = w_in[i_in + 1]
    elif chr(mode_in) == "l":  # linear
        a_out[0] = t1 * w_in[i_in] + t0 * w_in[i_in + 1]
    elif chr(mode_in) == "h":  # Cubic hermite
        m0 = w_in[1] - w_in[0] if i_in == 0 else (w_in[i_in + 1] - w_in[i_in - 1]) / 2
        m1 = (
            w_in[-1] - w_in[-2]
            if i_in == len(w_in) - 2
            else (w_in[i_in + 2] - w_in[i_in]) / 2
        )
        a_out[0] = (
            (-2 * t1**3 + 3 * t1**2) * w_in[i_in]
            + (-2 * t0**3 + 3 * t0**2) * w_in[i_in + 1]
            - (t1**3 - t1**2) * m0
            + (t0**3 - t0**2) * m1
        )
    elif chr(mode_in) == "s":  # Cubic spline
        u = np.zeros(len(w_in))
        w2 = np.zeros(len(w_in))  # second derivative values

        for i in range(1, len(w_in) - 1):
            p = 0.5 * w2[i - 1] + 2
            w2[i] = -0.5 / p
            u[i] = w_in[i + 1] - 2 * w_in[i] + w_in[i - 1]
            u[i] = (3 * u[i] - 0.5 * u[i - 1]) / p
        for i in range(len(w_in) - 2, i_in - 1, -1):
            w2[i] = w2[i] * w2[i + 1] + u[i]

        a_out[0] = (
            t1 * w_in[i_in]
            + t0 * w_in[i_in + 1]
            + ((t1**3 - t1) * w2[i_in] + (t0**3 - t0) * w2[i_in + 1]) / 6.0
        )
    else:
        raise DSPFatal("Unrecognized interpolation mode")



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32[:], float32[:])",
        "void(float64[:], float64, float64[:], float64[:])",
    ],
    "(n), (), (), (m)",
    **nb_kwargs,
)
def presum(w_in: np.ndarray, do_norm: int, ps_fact: int, w_out: np.ndarray) -> None:
    """Presum the waveform.

    Combine bins in chunks of ``len(w_in) / len(w_out)``, which is hopefully an
    integer. If it isn't, then some samples at the end will be omitted.

    Parameters
    ----------
    w_in
        the input waveform.
    do_norm
        a flag for setting returned chunks are normalized

        * ``0``: chunks are not normalized
        * ``1``: chunks are normalized

    ps_fact
        the presum factor/rate, determined by `len(w_in) // len(w_out)`
    w_out
        the output waveform.
    """
    w_out[:] = np.nan
    ps_fact[0] = np.nan

    if np.isnan(w_in).any():
        return
    if do_norm not in [int(0), int(1)]:
        raise DSPFatal("do_norm type not found.")

    ps_fact[0] = int(len(w_in) // len(w_out))
    for i in range(0, len(w_out), 1):
        j0 = i * int(ps_fact[0])
        if do_norm == 1:
            w_out[i] = w_in[j0] / ps_fact[0]
        else:
            w_out[i] = w_in[j0]

        for j in range(j0 + 1, j0 + ps_fact[0], 1):
            if do_norm == 1:
                w_out[i] += w_in[j] / ps_fact[0]
            else:
                w_out[i] += w_in[j]
